"""A module of various methods used to estimate aerofoil parameters."""
import numpy as np
from scipy.integrate import simpson

from carpy.utility import Hint, cast2numpy, point_diff

__all__ = ["ThinAerofoil"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Support Functions and Classes
# ---------------------------------------------------------------------------- #

class AerofoilSolution(object):
    """
    Template object, of which solvers should aim to produce all attributes of.
    """
    _Cd: float
    _Cl: float
    _Clalpha: float
    _Cm: Hint.func
    _Cm_ac: float
    _x_ac: float
    _x_cp: float
    _alpha_zl: float

    def __str__(self):
        params = ["Cd", "Cl", "Clalpha", "Cm_ac", "x_ac", "x_cp", "alpha_zl"]

        return_string = f"{type(self).__name__}\n"
        for param in params:
            if hasattr(self, f"_{param}"):
                return_string += f">\t{param:<8} = {getattr(self, param)}\n"

        return return_string

    @property
    def Cd(self) -> float:
        """Sectional pressure drag coefficient."""
        return self._Cd

    @property
    def Cl(self) -> float:
        """Sectional lift coefficient."""
        return self._Cl

    @property
    def Clalpha(self) -> float:
        """Sectional lift-curve slope."""
        return self._Clalpha

    @property
    def Cm(self) -> Hint.func:
        """Sectional moment coefficient, at a given position along the chord."""
        return self._Cm

    @property
    def Cm_ac(self) -> float:
        """Sectional moment coefficient, at the aerofoil aerodynamic centre."""
        return self._Cm_ac

    @property
    def x_ac(self) -> float:
        """Chordwise position of the aerofoil's aerodynamic centre."""
        return self._x_ac

    @property
    def x_cp(self) -> float:
        """Chordwise position of the aerofoil's centre of pressure."""
        return self._x_cp

    @property
    def alpha_zl(self) -> float:
        """The aerofoil's angle (of attack) of zero-lift."""
        return self._alpha_zl


# ============================================================================ #
# Public classes
# ---------------------------------------------------------------------------- #

class ThinAerofoil(AerofoilSolution):
    """
    Thin and cambered aerofoil theory.

    For use when the following assumptions are taken:

    -   Flow is incompressible
    -   Flow is irrotational
    -   Flow is inviscid (no viscosity)
    -   Angle of attack is small
    -   Aerofoil thickness is small
    -   Camber is small
    -   Drag = 0

    References:
        -   Fundamentals of Aerodynamics 6th Ed. Chapter 4, John D. Anderson Jr.
    """

    def __init__(self, aerofoil, alpha: Hint.num, N: int = None):
        """
        Args:
            aerofoil: Aerofoil object.
            alpha: Angle of attack.
            N: Number of discretised points in the aerofoil's camber line.
        """
        # Recast as necessary
        N = 100 if N is None else int(N)

        # Fundamental equation of thin aerofoil theory, fourier cosine expansion
        xs = (1 - np.cos((theta := np.linspace(0, np.pi, N)))) / 2
        zs = np.interp(xs, *aerofoil._camber_points(step_target=0.05).T)
        dz_dx = point_diff(y=zs, x=xs)

        def f_A0(alpha: np.ndarray) -> np.ndarray:
            """
            Element A0 of the Fourier cosine solution to the fundamental
            equation of thin aerofoil theory.

            Args:
                alpha: Angle of attack.

            Returns:
                Coefficient A0 of the Fourier cosine solution.

            """
            A0 = alpha - 1 / np.pi * simpson(dz_dx, xs)
            return A0

        def f_An(n: Hint.nums) -> np.ndarray:
            """
            Element(s) An of the Fourier cosine solution to the fundamental
            equation of thin aerofoil theory.

            Args:
                n: The n'th element of the infinite fourier series, for n > 0.

            Returns:
                Coefficient(s) An of the Fourier cosine solution.

            """
            # Recast as necessary
            n = cast2numpy(n)

            An = np.zeros(n.shape)
            for i, ni in enumerate(n.flat):
                An.flat[i] = simpson(dz_dx * np.cos(ni * theta), theta)
            else:
                An *= 2 / np.pi
            return An

        self._aerofoil = aerofoil
        self._A0 = float(f_A0(alpha=alpha))
        self._f_An = f_An

        return

    @property
    def _Cd(self) -> float:
        """The sectional drag coefficient."""
        return 0

    @property
    def _Cl(self) -> float:
        """The sectional lift coefficient."""
        cl = float(self._Clalpha * (self._A0 + self._f_An(1) / 2))
        return cl

    @property
    def _Clalpha(self) -> float:
        """The sectional lift-curve slope."""
        a0 = 2 * np.pi
        return a0

    def _Cm(self, x: Hint.nums) -> np.ndarray:
        # Recast as necessary
        x = cast2numpy(x)

        Cm = self._Cm_ac - (x - 0.25) * self._Cl

        return Cm

    @property
    def _Cm_ac(self) -> float:
        """Moment coefficient at the aerodynamic centre."""
        A1, A2 = self._f_An([1, 2])
        cm_ac = float((np.pi / 4) * (A2 - A1))
        return cm_ac

    @property
    def _x_ac(self) -> float:
        """The chordwise position of the aerodynamic centre."""
        return 0.25

    @property
    def _x_cp(self) -> float:
        """The chordwise position of the centre of pressure."""
        A1, A2 = self._f_An([1, 2])
        xc_cp = 0.25 * (1 + np.pi / self._Cl * (A2 - A1))

        if xc_cp < 0:
            return np.nan

        return xc_cp

    @property
    def _alpha_zl(self) -> float:
        """The angle of attack that produces zero lift."""
        # Find points describing camber, and redistribute this in cosine spacing
        xcamber, ycamber = self._aerofoil._camber_points(step_target=0.05).T
        xs = (1 - np.cos((theta := np.linspace(0, np.pi, 100)))) / 2
        dz_dx = np.interp(xs, xcamber, point_diff(y=ycamber, x=xcamber))

        alpha_zl = (1 / np.pi) * simpson(dz_dx * (2 * xs), theta)
        return alpha_zl



"""Assorted constants used in physics or standards."""
import numpy as np

from carpy.utility._unitconversion import Quantity

__all__ = ["constants"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Constants of the Universe
# ---------------------------------------------------------------------------- #
# Constants from Wikipedia

class Physical(object):
    """
    Assorted physical constants. Return types of this class use the Quantity
    object to describe the appropriate units for the constant's value.

    Sourced from Wikipedia's list of physical constants.
    """

    @property
    def G(self) -> Quantity:
        """Newtonian constant of gravitation"""
        return Quantity(6.67430e-11, "m^{3} kg^{-1} s^{-2}")

    @property
    def c(self) -> Quantity:
        """Speed of light in vacuum"""
        return Quantity(299_792_458, "m s^{-1}")

    @property
    def h(self) -> Quantity:
        """Planck constant"""
        return Quantity(6.626_070_15e-34, "J Hz^{-1}")

    @property
    def hbar(self) -> Quantity:
        """Reduced Planck constant"""
        return self.h / (2 * np.pi)

    @property
    def mu_0(self) -> Quantity:
        """Vacuum magnetic permeability"""
        return Quantity(1.256_637_062_12e-6, "N A^{-2}")

    @property
    def Z_0(self) -> Quantity:
        """Characteristic impedance of vacuum"""
        return self.mu_0 * self.c

    @property
    def eps_0(self) -> Quantity:
        """Vacuum electric permittivity"""
        return 1 / (self.mu_0 * self.c ** 2)

    @property
    def k_e(self) -> Quantity:
        """Coulomb constant"""
        return 1 / (4 * np.pi * self.eps_0)

    @property
    def k_B(self) -> Quantity:
        """Boltzmann constant"""
        return Quantity(1.380_649e-23, "J K^{-1}")

    @property
    def sigma(self) -> Quantity:
        """Stefan-Boltzmann constant"""
        return np.pi ** 2 * self.k_B ** 4 / (60 * self.hbar ** 3 * self.c ** 2)

    @property
    def b(self) -> Quantity:
        """Wien wavelength displacement law constant"""
        return Quantity(2.897_771_955e10 - 3, "m K")

    @property
    def bprime(self) -> Quantity:
        """Wien frequency displacement law constant"""
        return Quantity(5.878_925_757e10, "Hz K^{-1}")

    @property
    def b_entropy(self) -> Quantity:
        """Wien entropy displacement law constant"""
        return Quantity(3.002_916_077, "m K")

    @property
    def e(self) -> Quantity:
        """Elementary charge"""
        return Quantity(1.602_176_634e-19, "C")

    @property
    def G_0(self) -> Quantity:
        """Conductance quantum"""
        return 2 * self.e ** 2 / self.h

    @property
    def R_K(self) -> Quantity:
        """von Klitzing constant"""
        return self.h * self.e ** -2

    @property
    def K_J(self) -> Quantity:
        """Josephson constant"""
        return 2 * self.e * self.h ** -1

    @property
    def Phi_0(self) -> Quantity:
        """Magnetic flux quantum"""
        return self.h / (2 * self.e)

    @property
    def alpha(self) -> Quantity:
        """Fine-structure constant"""
        return self.e ** 2 / (4 * np.pi * self.eps_0 * self.hbar * self.c)

    @property
    def m_e(self) -> Quantity:
        """Electron, proton, and neutron mass"""
        return Quantity(9.109_383_7015e-31, "kg")

    @property
    def m_p(self) -> Quantity:
        """Electron, proton, and neutron mass"""
        return Quantity(1.672_621_923_69e-27, "kg")

    @property
    def m_n(self) -> Quantity:
        """Electron, proton, and neutron mass"""
        return Quantity(1.674_927_498_04e-27, "kg")

    @property
    def m_mu(self) -> Quantity:
        """Muon mass"""
        return Quantity(1.883_531_627e-28, "kg")

    @property
    def m_tau(self) -> Quantity:
        """Tau particle mass"""
        return Quantity(3.167_54e-27, "kg")

    @property
    def m_t(self) -> Quantity:
        """Top quark mass"""
        return Quantity(3.0784e-25, "kg")

    @property
    def g_e(self) -> Quantity:
        """Electron g-factor"""
        return Quantity(-2.002_319_304_362_56)

    @property
    def g_mu(self) -> Quantity:
        """Muon g-factor"""
        return Quantity(-2.002_331_8418)

    @property
    def g_p(self) -> Quantity:
        """Proton g-factor"""
        return Quantity(5.585_694_6893)

    @property
    def mu_B(self) -> Quantity:
        """Bohr magneton"""
        return self.e * self.hbar / (2 * self.m_e)

    @property
    def mu_N(self) -> Quantity:
        """Nuclear magneton"""
        return self.e * self.hbar / (2 * self.m_p)

    @property
    def r_e(self) -> Quantity:
        """Classical electron radius"""
        return self.e ** 2 * self.k_e / (self.m_e * self.c ** 2)

    @property
    def sigma_e(self) -> Quantity:
        """Thomson cross section"""
        return (8 * np.pi / 3) * self.r_e ** 2

    @property
    def a_0(self) -> Quantity:
        """Bohr radius"""
        return self.hbar ** 2 / (self.k_e * self.m_e * self.e ** 2)

    @property
    def E_h(self) -> Quantity:
        """Hartree energy"""
        return self.alpha ** 2 * self.c ** 2 * self.m_e

    @property
    def Ry(self) -> Quantity:
        """Rydberg unit of energy"""
        return self.E_h / 2

    @property
    def R_infty(self) -> Quantity:
        """Rydberg constant"""
        return self.Ry / (self.h * self.c)

    @property
    def N_A(self) -> Quantity:
        """Avogadro constant"""
        return Quantity(6.022_140_76e23, "mol^{-1}")

    @property
    def R(self) -> Quantity:
        """Molar gas constant"""
        return self.N_A * self.k_B

    @property
    def F(self) -> Quantity:
        """Faraday constant"""
        return self.N_A * self.e

    @property
    def m_12C(self) -> Quantity:
        """Atomic mass of carbon-12"""
        return Quantity(1.992_646_879_92e-26, "kg")

    @property
    def M_12C(self) -> Quantity:
        """Molar mass of carbon-12"""
        return Quantity(11.999_999_9958e-3, "kg mol^{-1}")

    @property
    def m_u(self) -> Quantity:
        """Atomic mass constant, one-twelfth the mass of carbon-12"""
        return self.m_12C / 12

    @property
    def M_u(self) -> Quantity:
        """Molar mass constant, one-twelfth the molar mass of carbon-12"""
        return self.M_12C / 12


# ============================================================================ #
# Constants defined in Standards
# ---------------------------------------------------------------------------- #
# Constants from World Geodetic System

class WGS84(object):
    """
    Assorted constants and parameters derived from, or used in the construction
    of, the World Geodetic System (1984).
    """

    @property
    def a(self) -> Quantity:
        """Semi-major axis of the world ellipsoid"""
        return Quantity(6_378_137, "m")

    @property
    def omega(self) -> Quantity:
        """Angular velocity"""
        return Quantity(7.292115e-5, "rad s^{-1}")

    @property
    def GM(self) -> Quantity:
        """Geocentric gravitational constant (including mass of atmosphere)"""
        return Quantity(398600.5, "km^{3} s^{-2}")

    @property
    def Cbar2_0(self) -> Quantity:
        """
        Normalised second degree zonal harmonic coefficient of the gravitational
        potential.
        """
        return Quantity(-484.16685e-6)

    @property
    def inv_f(self) -> Quantity:
        """Flattening parameter of the world ellipsoid"""
        return Quantity(1 / 298.257_223_563)

    @property
    def e(self) -> Quantity:
        """Eccentricity parameter of the world ellipsoid"""
        return (self.inv_f * (2 - self.inv_f)) ** 0.5

    @property
    def b(self) -> Quantity:
        """Ellipsoidal polar radius"""
        return self.a * (1 - self.e ** 2) ** 0.5

    @property
    def r45(self) -> Quantity:
        """Radius at +45deg latitude on the world ellipsoid"""
        cos45squared = 0.5
        return (self.b ** 2 / (1 - self.e ** 2 * cos45squared)) ** 0.5


# ---------------------------------------------------------------------------- #
# Constants from ISO 2533 (1975)

class StandardSealevel(object):
    """
    A class organising standards of mean sea level conditions used in the ISO
    Standard Atmosphere, as it is described in ISO publication 2533:1975 first
    edition.
    """

    @property
    def g(self) -> Quantity:
        """Standard acceleration of free fall at 45deg 32' 33''"""
        return Quantity(9.80665, "m s^{-2}")

    @property
    def M(self) -> Quantity:
        """Air molar mass at sea level as obtained from the perfect gas law"""
        return Quantity(28.964_420, "kg kmol^{-1}")

    @property
    def p(self) -> Quantity:
        """Standard air pressure"""
        return Quantity(101.325e3, "Pa")

    @property
    def Rs(self) -> Quantity:
        """Specific gas constant"""
        return Quantity(287.052_87, "J K^{-1} kg^{-1}")

    @property
    def To(self) -> Quantity:
        """Thermodynamic ice-point temperature, at mean sea level"""
        return Quantity(273.15, "K")

    @property
    def T(self) -> Quantity:
        """Standard thermodynamic air temperature at mean sea level"""
        return Quantity(288.15, "K")

    @property
    def rho(self) -> Quantity:
        """Standard air density"""
        return Quantity(1.225, "kg m^{-3}")

    @property
    def gamma(self) -> Quantity:
        """
        Adiabatic index, the ratio of specific heat of air at constant pressure
        to its specific heat at constant volume.
        """
        return Quantity(1.4)

    @property
    def sigma(self) -> Quantity:
        """
        Effective collosion diameter of an air molecule, taken as constant with
        altitude.
        """
        return Quantity(0.365e-9, "m")

    @property
    def X(self) -> dict:
        """
        Dry clean air composition at sea level.

        Because Ozone (O3), Sulphur dioxide (SO2), Nitrogen dioxide (NO2) and
        Iodine (I2) quantities can vary from time to time or place to place,
        their contributions are omitted.
        """
        composition = dict([
            ("N2", 78.084), ("O2", 20.947_6), ("Ar", 0.934), ("CO2", 0.031_4),
            ("Ne", 1.818e-3), ("He", 524.0e-6), ("Kr", 114.0e-6),
            ("Xe", 8.7e-6), ("H2", 50.0e-6), ("N2O", 50.0e-6), ("CH4", 0.2e-3)
        ])
        return composition


class ISO2533(StandardSealevel):
    """
    A class organising standards used in the ISO Standard Atmosphere, as it is
    described in ISO publication 2533:1975 first edition.
    """

    @property
    def NA(self) -> Quantity:
        """Avogardro's constant, as adopted in 1961 by IUPAC"""
        return Quantity(602.257e24, "kmol^{-1}")

    @property
    def R(self) -> Quantity:
        """Universal gas constant"""
        return Quantity(8_314.32, "J K^{-1} kmol^{-1}")

    @property
    def S(self) -> Quantity:
        """
        Sutherland empirical coefficient S in equation for dynamic viscosity.
        """
        return Quantity(110.4, "K")

    @property
    def betaS(self) -> Quantity:
        """
        Sutherland empirical coefficient beta in equation for dynamic viscosity.
        """
        return Quantity(1.458e-6, "kg m^{-1} s^{-1} K^{-0.5}")

    @property
    def r(self) -> Quantity:
        """Nominal Earth's radius"""
        return Quantity(6_356_766, "m")


# ---------------------------------------------------------------------------- #
# Constants from the joint Astronautical Almanac


class AA21K(object):
    """
    Astronautical Almanac (2021) section K, as described by HM Nautical Almanac
    Office and United States National Observatory.

    Notes:
        https://aa.usno.navy.mil/downloads/publications/Constants_2021.pdf

    """

    @property
    def GM_E(self) -> dict:
        """
        Gravitational parameter of Earth, a.k.a geocentric gravitational
        constant.

        Returns:
            Geocentric gravitational constants for TCB (Barycentric Coordinate
                Time), TT (Terrestrial Time), TDB (Barycentric Dynamical Time).

        """
        GM_E = {
            "TCB": Quantity(3.986_004_418e14, "m^{3} s^{-2}"),
            "TT": Quantity(3.986_004_415e14, "m^{3} s^{-2}"),
            "TDB": Quantity(3.986_004_356e14, "m^{3} s^{-2}")
        }
        return GM_E

    @property
    def M_E(self) -> Quantity:
        """Mass of the Earth"""
        return Quantity(5.97_22e24, "kg")


# ---------------------------------------------------------------------------- #

class Standards(object):
    """A class structure for organising constants based on standards"""
    _aa21k = AA21K()
    _iso2553 = ISO2533()
    _ssl = StandardSealevel()
    _wgs84 = WGS84()

    @property
    def AA21_K(self) -> AA21K:
        """Constants defined in the Astronautical Almanac, Section K, 2021."""
        return self._aa21k

    @property
    def ISO_2533(self) -> ISO2533:
        """Conditions of the International Standard Atmosphere per ISO 2553."""
        return self._iso2553

    @property
    def SL(self) -> StandardSealevel:
        """Conditions at sea level in the standard atmosphere per ISO 2553."""
        return self._ssl

    @property
    def WGS84(self) -> WGS84:
        """Parameters of the World Geodetic System 1984."""
        return self._wgs84


# ============================================================================ #
# Useful guesses for values in materials, from wherever on the internet
# ---------------------------------------------------------------------------- #

class esg_roughness(object):
    """
    Values from DATCOM Table 4.1.5.1-A and S. Gudmundsson General Aviation
    Aircraft Design (methods on estimating skin friction).
    """

    @property
    def camopaint_careful(self) -> Quantity:
        """Carefully applied camouflage paint on aluminium."""
        return Quantity(10.16, "um")

    @property
    def camopaint_production(self) -> Quantity:
        """Production-level quality application of camouflage paint."""
        return Quantity(30.48, "um")

    @property
    def castiron_production(self) -> Quantity:
        """Production-level quality of cast iron surface."""
        return Quantity(254, "um")

    @property
    def composite_molded(self) -> Quantity:
        """Smooth, molded composite surface."""
        return Quantity(0.518, "um")

    @property
    def mattepaint_careful(self) -> Quantity:
        """Careful application of matte paint."""
        return Quantity(6.350, "um")

    @property
    def metal_dipgalvanised(self) -> Quantity:
        """Dip galvanised metal."""
        return Quantity(152.4, "um")

    @property
    def metal_polished(self) -> Quantity:
        """Carefully polished metal surface."""
        return Quantity([0.508, 2.032], "um")

    @property
    def sheetmetal_polished(self) -> Quantity:
        """Carefully polished sheet metal surface."""
        return Quantity(1.524, "um")

    @property
    def sheetmetal_production(self) -> Quantity:
        """Production-level quality of sheet metal surface."""
        return Quantity(4.064, "um")

    @property
    def wood_polished(self) -> Quantity:
        """Carefully polished wooden surface."""
        return self.metal_polished


# ---------------------------------------------------------------------------- #

class Material(object):
    """Library of aeronautical material references"""

    _esg_roughness = esg_roughness()

    @property
    def esg_roughness(self) -> esg_roughness:
        """
        Returns a dictionary of representative values for (wing and aerofoil)
        surface roughness (equivalent sand-grain roughness).

        Values from DATCOM Table 4.1.5.1-A and S. Gudmundsson General Aviation
        Aircraft Design (methods on estimating skin friction).
        """
        return self._esg_roughness


# ============================================================================ #
class Constants(object):
    """A class structure for organising assortments of constants"""

    _material = Material()
    _physical = Physical()
    _standard = Standards()

    @property
    def MATERIAL(self) -> Material:
        """A library of assorted reference material properties."""
        return self._material

    @property
    def PHYSICAL(self) -> Physical:
        """Physical constants of the Universe."""
        return self._physical

    @property
    def STANDARD(self) -> Standards:
        """Constants defined by standards."""
        return self._standard


# Allow the below object to be imported and used by users
constants = Constants()



import unittest

from carpy.aerodynamics.aerofoil import NewAerofoil, ThinAerofoil


class Profiles(unittest.TestCase):
    """Methods to test aerofoil section generation and file parsing."""

    def test_generateNACA(self):
        # Four digit series
        n0012 = NewAerofoil.from_method.NACA("0012")
        n2412 = NewAerofoil.from_method.NACA("2412")
        n2412_63 = NewAerofoil.from_method.NACA("2412-63")
        # Five digit series
        n23012 = NewAerofoil.from_method.NACA("23012")
        n23012_45 = NewAerofoil.from_method.NACA("23012-45")
        n44112 = NewAerofoil.from_method.NACA("44112")
        # 16-series
        n16_012 = NewAerofoil.from_method.NACA("16-012")
        n16_912_3 = NewAerofoil.from_method.NACA("16-912,a=0.3")
        return


class ThinAerofoilTheory(unittest.TestCase):
    """Methods to test thin aerofoil theory."""

    def test_liftcurveslope(self):
        """Thin aerofoil theory suggests ideal lift slope of 2 pi."""
        flatplate = NewAerofoil.from_method.NACA("0001")
        solution = ThinAerofoil(aerofoil=flatplate, alpha=0)
        Clalpha = solution.Clalpha
        self.assertAlmostEqual(Clalpha, 2 * 3.1415926535, places=5)
        return


if __name__ == '__main__':
    unittest.main()



"""Tests for miscellaneous helper functions in carpy."""
import unittest

import numpy as np

from carpy.utility import (
    Hint,
    cast2numpy,
    GetPath
)


class HintsAndCasting(unittest.TestCase):

    def test_hint(self):
        testcases = (
            (Hint.iter, (
                (1, 2, 3), [1, 2, 3], np.arange(3)
            )),
            (Hint.num, (
                1, 2.0, np.int32(5), np.float32(6.0)
            )),
            (Hint.nums, (
                (1, 2, 3), [1, 2, 3], np.arange(3),
                1, 2.0, np.int32(5), np.float32(6.0)
            )),
            (Hint.func, (
                lambda x: print(f"You'll never take me alive! (also {x})"),
                np.radians, print
            )),
            (Hint.set, (
                {1, 2, 3}, frozenset((1, 2, 3))
            ))
        )
        for hint, tests in testcases:
            # Unions of multiple types
            if type(hint) == "typing._UnionGenericAlias":
                for obj in tests:
                    self.assertTrue(
                        expr=isinstance(obj, hint.__args__),
                        msg=f"Couldn't assert {obj} is instance of {hint}"
                    )
            elif hasattr(hint, "__args__"):
                for obj in tests:
                    self.assertTrue(
                        expr=isinstance(obj, hint.__args__),
                        msg=f"Couldn't assert {obj} is instance of {hint}"
                    )
            else:
                for obj in tests:
                    self.assertTrue(
                        expr=isinstance(obj, hint),
                        msg=f"Couldn't assert {obj} is instance of {hint}"
                    )
        return

    def test_cast2numpy(self):
        # Hashables
        testcases = {
            -1: np.array([-1]), 0: np.array([0]), 1: np.array([1]),
            0.0: np.array([0.0]),
            1 + 2j: np.array([1 + 2j]),
            np.int32(3): np.array([3]),
            np.float32(4): np.array([4]),
            (2, 3): np.array([2, 3]),
            frozenset((2, 3)): np.array([2, 3]),
        }
        # Test hashables
        for testcase, goldresult in testcases.items():
            casting = cast2numpy(testcase)
            # Check the value came through alright
            self.assertTrue(
                expr=all(casting == goldresult),
                msg=f"Failed to cast {testcase} with correct value"
            )
            # Check that the datatype matches
            self.assertIsInstance(
                obj=casting,
                cls=type(goldresult),
                msg=f"Failed to cast {testcase} with correct type"
            )

        # Unhashables
        testcases = [
            ([2, 3], np.array([2, 3])), ({2, 3}, np.array([2, 3])),
            ({"a": 1, "b": {"c": 2}},
             {"a": np.array([1]), "b": {"c": np.array([2])}}),
            (np.array(-6), np.array([-6]))
        ]

        # Test unhashables
        for testcase, goldresult in testcases:
            casting = cast2numpy(testcase)
            # Check the value came through alright
            if not isinstance(testcase, dict):
                self.assertTrue(
                    expr=all(casting == goldresult),
                    msg=f"Failed to cast {testcase} with correct value"
                )
            else:
                self.assertTrue(
                    expr=(casting == goldresult),
                    msg=f"Failed to cast {testcase} with correct value"
                )
            # Check that the datatype matches
            self.assertIsInstance(
                obj=casting,
                cls=type(goldresult),
                msg=f"Failed to cast {testcase} with correct type"
            )

        return


class Pathing(unittest.TestCase):

    @staticmethod
    def test_getpath():
        GetPath.library()
        GetPath.localpackage()
        return


if __name__ == '__main__':
    unittest.main()



"""Tests for airfields, airports, and runways."""
import unittest

from carpy.environment import Airport, Runway


class RunwayTest(unittest.TestCase):
    """Test that parameters of runways are established successfully."""

    def test_attributes(self):
        """Test returned values of attributes of the runway."""

        # Instantiate a runway (CYVR 08L/26R)
        myrunway = Runway(234512)

        self.assertEqual(myrunway.airport_ident, "CYVR")
        self.assertEqual(myrunway.length.to("ft"), 9940)
        self.assertEqual(myrunway.width.to("ft"), 200)
        self.assertTrue(myrunway.lighted is True)
        self.assertTrue(myrunway.closed is False)
        self.assertEqual(myrunway.le_ident, "08L")
        self.assertAlmostEqual(myrunway.le_latitude_deg, 49.205000, places=3)
        self.assertAlmostEqual(myrunway.le_longitude_deg, -123.201000, places=3)
        self.assertEqual(myrunway.le_elevation.to("ft"), 13)
        self.assertEqual(myrunway.le_heading_degT, 100)
        self.assertEqual(myrunway.le_displaced_threshold, 0)
        self.assertEqual(myrunway.he_ident, "26R")
        self.assertAlmostEqual(myrunway.he_latitude_deg, 49.200600, places=3)
        self.assertAlmostEqual(myrunway.he_longitude_deg, -123.160000, places=3)
        self.assertEqual(myrunway.he_elevation.to("ft"), 9)
        self.assertEqual(myrunway.he_heading_degT, 280)
        self.assertEqual(myrunway.he_displaced_threshold, 0)

        return


class AirportTest(unittest.TestCase):

    def test_attributes(self):
        """Test returned values of attributes of the runway."""

        # Instantiate an airport (EGLL, London Heathrow)
        myairport = Airport("EGLL")

        # Test loaded airport attributes
        self.assertEqual(myairport.ident, "EGLL")
        self.assertEqual(myairport.name, "London Heathrow Airport")
        self.assertAlmostEqual(myairport.latitude_deg, 51.470600, places=3)
        self.assertAlmostEqual(myairport.longitude_deg, -0.461941, places=3)
        self.assertEqual(myairport.elevation.to("ft"), 83)

        # Test we get the correct runways built
        for runway in myairport.runways:
            self.assertEqual(runway.airport_ident, myairport.ident)

        return




"""Tests for carpy methods designed to make things prettier."""
import unittest

from carpy.utility import Unicodify, Pretty


class MakePrettyMethods(unittest.TestCase):

    def test_pretty_temperature(self):
        test_input = [(10., None), (10.0, None), (10.60, None), (10.606, None),
                      (10.606, "degF"), (10.32, "K"), (-3, "degR")]
        gold_output = ['+10°C', '+10°C', '+10.6°C', '+10.606°C',
                       '+10.606°F', '+10.32 K', '-3°R']
        for i, (temperature, unit) in enumerate(test_input):
            result = Pretty.temperature(temperature, unit)
            self.assertEqual(first=result, second=gold_output[i])
        return


class UnicodeScripting(unittest.TestCase):

    def test_shallowscript(self):
        result = Unicodify.mathscript_safe("g_{0} = 9.81 ms^{2}")
        gold_output = "g₀ = 9.81 ms²"
        self.assertEqual(first=result, second=gold_output)
        return


if __name__ == '__main__':
    unittest.main()



"""A package of utilities for carpy and its users."""
from ._constants import *
from ._graphtheory import *
from ._maths import *
from ._miscellaneous import *
from ._unitconversion import *
from ._vanity import *



"""Methods for modelling the 3d geometry of a fuselage."""
import numpy as np
import streamlit as st


class SuperEllipse(object):

    def __init__(self):
        # Position offset
        self._posX = 0
        self._posY = 0

        # Shape parameters (default for a circle)
        self._m = 2  # Default for an ellipse-type shape
        self._n = 1  # Default for an ellipse-type shape
        self._a = 1
        self._b = 1
        return

    def plot(self, N: int):
        # T is a parameter with no physical interpretation
        t = np.linspace(0, np.pi / 2, N)

        x = self._a * np.cos(t) ** (2 / self._m)
        y = self._b * np.sin(t) ** (2 / self._n)

        from matplotlib import pyplot as plt

        plt.plot(x, y)
        plt.plot(x, -y)
        plt.plot(-x, -y)
        plt.plot(-x, y)
        plt.show()


if __name__ == "__main__":
    # import streamlit as st

    age = st.slider('How old are you?', 0, 130, 25)
    st.write("I'm ", age, 'years old')

    # Streamlit runs from the terminal. Have to do: streamlit "path-to-file.py"




"""
A module of tools related to the access of various properties and attributes
of the chemical elements.
"""
from functools import cached_property
import re
from typing import Union

import numpy as np
import periodictable as pt

from carpy.utility import Hint, Unicodify

__all__ = ["Elements"]
__author__ = "Yaseen Reza"

symbols_alphabetical = set(
    x for x in dir(pt)
    if isinstance(getattr(pt, x), pt.core.Element)  # Ignore isotopes
    and re.match("[A-Z][a-z]{0,2}", x)
)


# noinspection PyTypeChecker
class ElementTools(object):
    """
    A class of tools related to the access of various properties and attributes
    of the chemical elements.
    """

    @staticmethod
    def get(symbol: str) -> pt.core.Element:
        """
        Obtain an element object from the periodictable module.

        Args:
            symbol: Two-character symbol representing the element.

        Returns:
            periodictable.core.Element.

        """
        if symbol in symbols_alphabetical:
            return getattr(pt, symbol)
        raise ValueError(f"Did not recognise '{symbol=}' as a valid element.")

    @property
    def regex(self) -> re.Pattern:
        """Regular expression object for identifying elements in a string."""
        elements_by_len = sorted(self.by_Z, key=lambda x: len(x.symbol))[::-1]
        elements_pattern = f"(?:{'|'.join(x.symbol for x in elements_by_len)})"
        return re.compile(elements_pattern)

    @property
    def nobles(self) -> tuple[pt.core.Element]:
        """Noble gases."""
        return tuple(getattr(pt, x) for x in "He,Ne,Ar,Kr,Xe,Rn,Og".split(","))

    @property
    def lanthanides(self) -> tuple[pt.core.Element]:
        """Lanthanide series elements."""
        return tuple(x for x in self.by_Z if x.number in range(57, 72))

    @property
    def actinides(self) -> tuple[pt.core.Element]:
        """Actinide series elements."""
        return tuple(x for x in self.by_Z if x.number in range(89, 104))

    def spdf(self, x, /, abbreviate: bool = None, pretty: bool = None) -> str:
        """
        Returns the electron configuration for a given element.

        Args:
            x (int, str): Atomic mass unit or symbol of element from which
                configuration is desired.
            abbreviate: True or False flag for abbreviating the output with
                representative noble gases. Optional, defaults to True.
            pretty: True or False flag, adds superscript unicode to the output.
                Optional, defaults to False.

        Returns:
            String representing the spdf orbital arrangement of electrons.

        """
        # Recast as necessary
        if isinstance(x, (pt.core.Element, pt.core.Isotope)):
            Z = int(x.number)
        elif isinstance(x, Hint.num.__args__):
            Z = int(x)
        elif isinstance(x, str):
            matches = self.regex.findall(x)
            if len(matches) > 1:
                raise ValueError(f"Got too many elements (got {matches=})")
            elif len(matches) == 0:
                raise ValueError(f"Did not recognise element={x}")
            Z = int(self.get(matches[0]).number)
        else:
            raise ValueError(f"Did not recognise {x} as a valid argument")
        abbreviate = True if abbreviate is None else abbreviate
        pretty = False if pretty is None else pretty

        # Use a dictionary to instantly return electron configurations
        def uncontract_noble(_config: str):
            """Convert abbreviated spdf representations into verbose forms."""
            config_nobles = {
                "[He]": "1s2",
                "[Ne]": "[He] 2s2 2p6",
                "[Ar]": "[Ne] 3s2 3p6",
                "[Kr]": "[Ar] 4s2 3d10 4p6",
                "[Xe]": "[Kr] 5s2 4d10 5p6",
                "[Rn]": "[Xe] 6s2 4f14 5d10 6p6",
                "[Og]": "[Rn] 7s2 5f14 6d10 7p6"
            }
            while "[" in _config:
                for key, value in config_nobles.items():
                    if _config.startswith(key):
                        _config = _config.replace(key, value)
                        break
                else:
                    raise StopIteration(f"Invalid config string '{_config=}'")
            return _config

        config = {
            # These are exceptions to the rule on the filling order of e- shells
            24: "[Ar] 3d5 4s1", 29: "[Ar] 3d10 4s1",
            41: "[Kr] 5s1 4d4", 42: "[Kr] 5s1 4d5",
            44: "[Kr] 5s1 4d7", 45: "[Kr] 5s1 4d8",
            46: "[Kr] 4d10", 47: "[Kr] 5s1 4d10",
            57: "[Xe] 6s2 5d1", 58: "[Xe] 6s2 4f1 5d1",
            64: "[Xe] 6s2 4f7 5d1", 78: "[Xe] 6s1 4f14 5d9",
            79: "[Xe] 6s1 4f14 5d10", 89: "[Rn] 7s2 6d1",
            90: "[Rn] 7s2 6d2", 91: "[Rn] 7s2 5f2 6d1",
            92: "[Rn] 7s2 5f3 6d1", 93: "[Rn] 7s2 5f4 6d1",
            96: "[Rn] 7s2 5f7 6d1", 103: "[Rn] 7s2 5f14 7p1",
            # Noble gases should also be returned instantly
            **{element.number: f"[{element.symbol}]" for element in self.nobles}
        }.get(Z, None)

        if config is not None:
            if abbreviate is False:
                config = uncontract_noble(config)
            if pretty is True:
                config = Unicodify.superscript_trailingnums(config)
            return config

        # If we haven't returned, that's because we need to solve for the config
        lesser_nobles = tuple(x for x in self.nobles if x.number < Z)
        if any(lesser_nobles):
            Z -= lesser_nobles[
                -1].number  # Consume Z by greatest noble less than Z

        electron_shell_def = {"s": 2, "f": 14, "d": 10, "p": 6}
        electron_fill_order = {
            1: ("1s2",),
            2: ("2s2", "2p6"),
            3: ("3s2", "3p6"),
            4: ("4s2", "3d10", "4p6"),
            5: ("5s2", "4d10", "5p6"),
            6: ("6s2", "5d1", "4f14", "5d9", "6p6"),
            7: ("7s2", "6d1", "5f14", "6d9", "7p6")
        }

        for _, (period, shells) in enumerate(electron_fill_order.items()):
            if period <= len(lesser_nobles):
                continue  # For every noble configuration contained, skip period

            # Initialise an empty period so we can build up the orbitals
            period_sfdp = {k: 0 for (k, _) in electron_shell_def.items()}
            for _, shell in enumerate(shells):
                # 3d10 -> orbital = 'd', capacity = 10
                orbital, capacity = shell[1], int(shell[2:])

                # Distribute Z among orbitals
                dZ = min(Z, capacity)
                Z -= dZ
                period_sfdp[orbital] += dZ

            if any(lesser_nobles):
                config = [f"[{lesser_nobles[-1].symbol}]"]
            else:
                config = [""]

            # Periods fill d- and f- orbitals 1 and 2 periods lower respectively
            p_offset = {"s": 0, "p": 0, "d": 1, "f": 2}
            for _, (orbital, valence) in enumerate(period_sfdp.items()):
                if valence == 0:
                    continue
                config.append(
                    f"{period - p_offset[orbital]}{orbital}{valence}")
            break  # Do not ascend to the next period, exit the for-loop now

        # Finish up
        config = " ".join(config)
        if abbreviate is False:
            config = uncontract_noble(config)
        if pretty is True:
            config = Unicodify.superscript_trailingnums(config)
        return config

    def valence(self, element: Union[pt.core.Element, str]) -> int:
        """
        Returns the number of electrons in the highest principle energy shell.

        Args:
            element: Symbol of the element to query in the periodic table.

        Returns:
            int: The number of valence electrons in the element.

        """
        # Take the electronic configuration
        e_config = self.spdf(element, abbreviate=False, pretty=False)
        periods, subshells, electrons = zip(*[
            (int(p), s, int(n))
            for (p, s, n) in re.findall(r"(\d)([spdf])(\d+)", e_config)
        ])
        # Identify the greatest period that appears and work out valence
        arguments = np.argwhere(np.array(periods) == max(periods))[:, 0]
        valence = np.sum(np.array(electrons)[arguments])
        return valence

    def locate(self, element: Union[pt.core.Element, str]) -> tuple:
        """
        Returns the location of the element in the periodic table.

        Args:
            element: Symbol of the element to locate in the periodic table.

        Returns:
            tuple: The period (row number) and group (column number) at which
                the element resides.

        """
        # Take the electronic configuration
        e_config = self.spdf(element, abbreviate=False, pretty=False)
        periods, subshells, electrons = zip(*[
            (int(p), s, int(n))
            for (p, s, n) in re.findall(r"(\d)([spdf])(\d+)", e_config)
        ])
        # Identify the greatest period that appears and work out # of outer elec
        argument = np.argmax(periods)
        period = periods[argument]
        period_electrons_spd = sum([
            x for (i, x) in enumerate(electrons[argument:])
            if subshells[argument:][i] != "f"
        ])

        # Special corrections
        # ... because Paladium is annoying like that
        if sum(electrons) == 46:
            period, group = (5, 10)
        # ... period 1 is missing p, d, and f orbitals
        elif period == 1 and period_electrons_spd == 2:
            group = period_electrons_spd + 16
        # ... periods 2 and 3 are missing d and f orbitals
        elif period in (2, 3) and period_electrons_spd > 2:
            group = period_electrons_spd + 10
        else:
            group = period_electrons_spd

        return period, group

    @property
    def by_alpha(self) -> tuple[pt.core.Element]:
        """Elements in alphabetical order."""
        return tuple(getattr(pt, x) for x in symbols_alphabetical)

    @property
    def by_Z(self) -> tuple[pt.core.Element]:
        """Elements in order of increasing atomic mass."""
        return tuple(sorted(self.by_alpha, key=lambda element: element.number))

    @cached_property
    def by_pseudoChi(self) -> tuple[pt.core.Element]:
        """Elements in recommended order for inorganic chemical nomenclature."""
        locations = {x: self.locate(x) for x in self.by_Z}
        rlocations = {v: k for (k, v) in locations.items()}

        # Step 1: Taking atomic number sequence, reverse sort by group
        def sort_stage1(key):
            """Score elements in a way that forces reverse group order."""
            period, group = self.locate(key)
            score1 = (18 - group) * 10
            score2 = period
            return score1 + score2

        sequence = sorted(locations, key=sort_stage1)

        # Step 2: Eliminate Lanthanides, Actinides, and reposition Nobles
        sequence = [
            x for x in sequence
            if not (x in self.lanthanides + self.actinides + self.nobles)
        ]
        sequence += list(self.nobles)

        # Step 3: Re-introduce Lanthanides and Actinides before Beryllium
        beryllium = rlocations.get((2, 2))
        beryllium_i = sequence.index(beryllium)
        [sequence.insert(beryllium_i, x) for x in self.actinides[::-1]]
        [sequence.insert(beryllium_i, x) for x in self.lanthanides[::-1]]

        # Step 4: Hydrogen is moved to precede Nitrogen
        hydrogen = rlocations.get((1, 1))
        hydrogen_i = sequence.index(hydrogen)
        del sequence[hydrogen_i]
        nitrogen = rlocations.get((2, 15))
        nitrogen_i = sequence.index(nitrogen)
        sequence.insert(nitrogen_i, hydrogen)

        return tuple(sequence)


# Destructive initialisation, no one will ever want the class version of this
Elements = ElementTools()



"""Methods for prescribing geometry the indexing of 'stations'."""
import numpy as np

from carpy.utility import Hint, cast2numpy, collapse_array

__all__ = ["DiscreteIndex", "ContinuousIndex"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Support functions
# ---------------------------------------------------------------------------- #

def validate_keys(*keys: Hint.any) -> tuple[bool, ...]:
    """
    Assert that arguments (for station numbers) are all numeric.

    Args:
        *keys: Dictionary keys to check.

    Returns:
        A tuple the same length as the *keys argument, true if the key is valid
         (numeric in nature), and false otherwise.

    Notes:
        Be careful - any() or all() functions should be used to determine the
            appropriate course of action after evaluation with this function.

    """
    valid_keys = tuple([
        True if isinstance(x, Hint.num.__args__) else False
        for x in keys
    ])
    return valid_keys


def bad_key(key) -> None:
    """Raise an error, and tell the user what was wrong."""
    errormsg = f"Station numbering {key=} is not numeric in nature!"
    raise ValueError(errormsg)


def sort_dict(dictionary: dict) -> dict:
    """Sort dictionaries by their keys."""
    sorted_dict = dict(sorted(dictionary.items(), key=lambda x: x[0]))
    return sorted_dict


def sort_kvs(nestedlist: list) -> list:
    """Sort results of a 'list(dict.items())' call by the dictionary keys."""
    sorted_list = sorted(nestedlist, key=lambda x: x[0])
    return sorted_list


def parse_slice(keys: np.ndarray, slice_object: slice) -> tuple:
    """Extract customised keys bounded by a slice object."""
    # Define the starting key, if it is missing
    if (start := slice_object.start) is None:
        start = keys[0]

    # If stop key is missing, use <= (get everything). Else, < (up to stop key)
    if (stop := slice_object.stop) is None:
        stop = keys[-1]
        realkeys = keys[(start <= keys) & (keys <= stop)]
    else:
        realkeys = keys[(start <= keys) & (keys < stop)]

    # If no step is given, just return the relevant keys
    if (step := slice_object.step) is None:
        return realkeys

    # Else a "step" is given, create a linear spacing of keys with 'step' elems.
    # noinspection PyArgumentList
    return np.linspace(keys.min(), keys.max(), step)


# ============================================================================ #
# Classes for consistent indexing behaviour
# ---------------------------------------------------------------------------- #
class TransparentArray(list):
    """
    Class for broadcasting get and set attributes to items *within* an array
    (instead of trying to access attributes of the array itself).
    """

    def __getattr__(self, item):
        return [getattr(x, item) for x in self]

    def __setattr__(self, name, value):
        if (n := len(self)) == 0:
            raise ValueError(f"Couldn't broadcast {value} to empty slice")

        # Broadcast assignment
        values = np.broadcast_to(value, n)
        [setattr(x, name, y) for (x, y) in np.broadcast(self, values)]
        return


class DiscreteIndex(dict):
    """
    Template class for Station-based geometry numbering systems, with linear
    interpolation between a discrete number of prescribed geometries.
    """

    def __init__(self, mapping: dict = None, /, **kwargs):
        """
        Args:
            mapping: Initial mapping of keys and values.
            **kwargs: Updates to apply to the mapping.
        """
        # Recast as necessary
        mapping = dict() if mapping is None else mapping
        if kwargs:
            mapping.update(dict([(k, v) for (k, v) in kwargs.items()]))

        # Validate mapping (keys)
        remapped = dict()
        for (key, value) in mapping.items():

            if validate_keys(key)[0]:
                # Ensure key maps as float for consistency of operation
                remapped[float(key)] = value
                continue

            bad_key(key)  # <-- Raise error on bad key

        # Superclass call (with sorted version of self)
        super().__init__(sort_dict(remapped))
        return

    def __setitem__(self, key, value):

        if validate_keys(key):
            # Record current entries of the dictionary, and clear self
            items = list(self.items())
            self.clear()

            # Ensure key maps as float for consistency of operation
            items.append((float(key), value))

            # Sort entries, and place them back into self
            items = sort_kvs(items)
            [super(DiscreteIndex, self).__setitem__(*kv) for kv in items]
            return

        elif isinstance(key, slice):
            raise NotImplementedError("Cannot set using slice object!")
        #     # Record current entries of the dictionary, and clear self
        #     items = list(self.items())
        #     self.clear()
        #
        #     # Ensure key maps as float for consistency of operation
        #     items.append((float(key), value))

        bad_key(key)  # <-- Raise error on bad key

    def __getitem__(self, key):

        # Need to get the key if it exists, else interpolate the data...
        keys, vals = zip(*self.items())
        keys, vals = map(cast2numpy, [keys, vals])  # Make indexable

        # If slice object, find the keys being referenced - else, only one key
        if isinstance(key, slice):
            keys2get = parse_slice(keys, slice_object=key)
            sliced = True
        else:
            keys2get = cast2numpy(key)  # Make iterable
            sliced = False

        vals2rtn = list()
        for key in keys2get:
            # If the key exists or is out of bounds, take closest matching key
            if key in keys:
                vals2rtn.append(super().__getitem__(key))
            elif key > max(keys):
                vals2rtn.append(super().__getitem__(keys[-1]))
            elif key < min(keys):
                vals2rtn.append(super().__getitem__(keys[0]))
            # Otherwise, we have to interpolate between keys in self
            else:
                # Find bounding keys
                keyidx_hi = np.argmax(key < keys)
                key_hi = keys[keyidx_hi]
                key_lo = keys[keyidx_hi - 1]

                # Linearly interpolate contribution of values at the keys
                weight = (key - key_lo) / (key_hi - key_lo)
                val_hi = super().__getitem__(key_hi)
                val_lo = super().__getitem__(key_lo)
                val_new = val_hi * weight + val_lo * (1 - weight)
                try:
                    # If the object allows new attributes, record parents
                    val_new._parents = (key_lo, key_hi)
                except AttributeError as _:
                    pass  # Couldn't assign attribute '_parents' to this object
                vals2rtn.append(val_new)

        # If sliced, return an array
        if sliced is True:
            return TransparentArray(vals2rtn)
        return collapse_array(TransparentArray(vals2rtn))


class ContinuousIndex(object):
    """
    Template class for Station-based geometry numbering systems. Station-wise
    geometry is prescribed procedurally, and so the set of accessible stations
    is continuous.
    """

    def __init__(self, procedure: Hint.func):
        self._procedure = procedure
        return

    def __setitem__(self, key, value):
        errormsg = (
            f"Procedurally generated stations cannot be overridden at discrete "
            f"points. Perhaps you meant to use this on a subclass of "
            f"{DiscreteIndex.__name__} (instead of "
            f"{ContinuousIndex.__name__})"
        )
        raise AttributeError(errormsg)

    def __getitem__(self, key):
        if isinstance(key, slice):
            errormsg = (
                f"Cannot obtain stations using index slicing (got {key}). This "
                f"is because the stations are defined continuously with a "
                f"procedure (and are therefore not iterables that can be "
                f"sliced). Use discrete station positions instead."
            )
            raise LookupError(errormsg)

        # Recast as necessary
        keys2get = cast2numpy(key)
        vals2rtn = [self._procedure(x) for x in keys2get]

        return vals2rtn[0] if len(vals2rtn) == 1 else tuple(vals2rtn)

# ============================================================================ #
# Classes for Stations
# ---------------------------------------------------------------------------- #

# Each station numbering type will have a different set of unique properties
# for example, fuselage geometry are always oriented with the x axis,
# buttock with y, etc.
# Fuselage: StationsLinear  # FS
# ButtockLine: StationsLinear  # BL
# WaterLine: StationsLinear  # WL
# Aileron: StationsLinear  # AS
# Flap: StationsLinear  # KS
# Nacelle: StationsLinear  # NC
# Wing: StationsLinear  # WS
# HorizontalStabiliser: StationsLinear  # HSS
# VerticalStabiliser: StationsLinear  # VSS
# Powerplant: StationsLinear  # PPS



"""
Vortex-source panel method for the pressure distribution around an aerofoil.

The code contained in this file is originally sourced from Aero Python (see
References section), and has been modified for improved speed and compatibility.

References:
    -   Barba, Lorena A., and Mesnard Olivier (2019). Aero Python: classical
        aerodynamics of potential flow using Python. Journal of Open Source
        Education, 2(15), 45, https://doi.org/10.21105/jose.00045.
    -   NACA0012 validation case:
        https://turbmodels.larc.nasa.gov/naca0012_val.html

"""
from functools import cached_property

import numpy as np
import scipy.integrate as sint

from carpy.utility import Hint, cast2numpy
from ._solvers import AerofoilSolution

__all__ = ["InviscidPanelMethod"]
__author__ = "Yaseen Reza"


class Panel(object):
    """
    Contains information related to a panel.
    """

    def __init__(self, xa: Hint.num, ya: Hint.num, xb: Hint.num, yb: Hint.num):
        """
        Initialise a panel (aerodynamic wall).

        Sets the panels end-points and calculates various features of the panel.
        Initialises source-strength, tangential velocity and pressure
        coefficient to zero.

        Args:
            xa: x-coordinate of the first end-point.
            ya: y-coordinate of the first end-point.
            xb: x-coordinate of the second end-point.
            yb: y-coordinate of the second end-point.
        """

        self.xa, self.ya = xa, ya  # panel starting-point
        self.xb, self.yb = xb, yb  # panel ending-point

        self.xc, self.yc = (xa + xb) / 2, (ya + yb) / 2  # panel center
        self.length = np.sqrt((xb - xa) ** 2 + (yb - ya) ** 2)  # panel length

        # orientation of panel (angle between x-axis and panel's normal)
        if xb - xa <= 0.0:
            self.beta = np.arccos((yb - ya) / self.length)
        else:
            self.beta = np.pi + np.arccos(-(yb - ya) / self.length)

        # panel location
        if self.beta <= np.pi:
            self.loc = 'upper'  # upper surface
        else:
            self.loc = 'lower'  # lower surface

        self.sigma = 0.0  # source strength
        self.vt = 0.0  # tangential velocity
        self.cp = 0.0  # pressure coefficient
        return

    @cached_property
    def cos_beta(self):
        """Pre-computed and cached result of the cosine of panel angle beta."""
        return np.cos(self.beta)

    @cached_property
    def sin_beta(self):
        """Pre-computed and cached result of the sine of panel angle beta."""
        return np.sin(self.beta)


def define_panels(aerofoil, N: int = None) -> np.ndarray:
    """
    Discretizes the geometry into panels using 'cosine' method.

    Args:
        aerofoil: The aerofoil geometry object to be broken into panels.
        N: Number of panels. Optional, defaults to 40.

    Returns:
        panels: 1D Numpy array of Panel objects.
            The list of panels.

    """
    # Recast as necessary
    points = aerofoil.section.geometry.points
    x, y = cast2numpy(points).T
    N = 40 if N is None else N

    R = (x.max() - x.min()) / 2.0  # circle radius
    x_center = (x.max() + x.min()) / 2.0  # x-coordinate of circle center

    theta = np.linspace(0.0, 2.0 * np.pi, N + 1)  # array of angles
    x_circle = x_center + R * np.cos(theta)  # x-coordinates of circle

    x_ends = np.copy(x_circle)  # x-coordinate of panels end-points
    y_ends = np.empty_like(x_ends)  # y-coordinate of panels end-points

    # extend coordinates to consider closed surface
    x, y = np.append(x, x[0]), np.append(y, y[0])

    def approx_bounded(a, x, b):
        """Function that evaluates whether x lies in the bounds [a, b]."""
        dir1 = (a <= x or np.isclose(a, x)) and (x <= b or np.isclose(x, b))
        dir2 = (b <= x or np.isclose(b, x)) and (x <= a or np.isclose(x, a))
        return dir1 | dir2

    # compute y-coordinate of end-points by projection
    infl = 0
    for i in range(N + 1):  # Iterate over panel end points that need computing
        while infl < len(x) - 1:  # Iterate over source coordinates
            if approx_bounded(x[infl], x_ends[i], x[infl + 1]):
                # If x_diff is zero, that means two coordinates with the same
                # abscissa. To not duplicate the 'y' value, 'I' must advance 1.
                if i > 0 and np.diff(x_ends)[i - 1] == 0:
                    infl += 1  # <-- implies next bounds set includes x_ends[i]
                break
            else:
                infl += 1
        # Linear interpolation
        a = (y[infl + 1] - y[infl]) / (x[infl + 1] - x[infl])
        b = y[infl + 1] - a * x[infl + 1]
        y_ends[i] = a * x_ends[i] + b

    # create panels
    panels = np.empty(N, dtype=object)
    for i in range(N):
        panels[i] = Panel(x_ends[i], y_ends[i], x_ends[i + 1], y_ends[i + 1])

    return panels


class Freestream:
    """
    Freestream conditions.
    """

    def __init__(self, u_inf=1.0, alpha=0.0):
        """
        Sets the freestream speed and angle.

        Parameters
        ----------
        u_inf: float, optional
            Freestream speed;
            default: 1.0.
        alpha: float, optional
            Angle of attack;
            default 0.0.
        """
        self.u_inf = u_inf
        self.alpha = float(alpha)

    @cached_property
    def cos_alpha(self):
        """Pre-computed and cached result of the cosine of angle of attack."""
        return np.cos(self.alpha)

    @cached_property
    def sin_alpha(self):
        """Pre-computed and cached result of the sine of angle of attack."""
        return np.sin(self.alpha)


def integral(x, y, panel, dxdk, dydk):
    """
    Evaluates the contribution from a panel at a given point.

    Parameters
    ----------
    x: float
        x-coordinate of the target point.
    y: float
        y-coordinate of the target point.
    panel: Panel object
        Panel whose contribution is evaluated.
    dxdk: float
        Value of the derivative of x in a certain direction.
    dydk: float
        Value of the derivative of y in a certain direction.

    Returns
    -------
    Contribution from the panel at a given point (x, y).
    """

    def integrand(s):
        """Given arc length s, find influence of panel on target point."""
        return (((x - (panel.xa - panel.sin_beta * s)) * dxdk +
                 (y - (panel.ya + panel.cos_beta * s)) * dydk) /
                ((x - (panel.xa - panel.sin_beta * s)) ** 2 +
                 (y - (panel.ya + panel.cos_beta * s)) ** 2))

    return sint.quad(integrand, 0.0, panel.length)[0]


def source_contribution_normal(panels):
    """
    Builds the source contribution matrix for the normal velocity.

    Parameters
    ----------
    panels: 1D array of Panel objects
        List of panels.

    Returns
    -------
    A: 2D Numpy array of floats
        Source contribution matrix.
    """
    A = np.empty((panels.size, panels.size), dtype=float)
    # source contribution on a panel from itself
    np.fill_diagonal(A, 0.5)
    # source contribution on a panel from others
    for i, panel_i in enumerate(panels):
        for j, panel_j in enumerate(panels):
            if i != j:
                A[i, j] = 0.5 / np.pi * integral(panel_i.xc, panel_i.yc,
                                                 panel_j,
                                                 panel_i.cos_beta,
                                                 panel_i.sin_beta)
    return A


def vortex_contribution_normal(panels):
    """
    Builds the vortex contribution matrix for the normal velocity.

    Parameters
    ----------
    panels: 1D array of Panel objects
        List of panels.

    Returns
    -------
    A: 2D Numpy array of floats
        Vortex contribution matrix.
    """
    A = np.empty((panels.size, panels.size), dtype=float)
    # vortex contribution on a panel from itself
    np.fill_diagonal(A, 0.0)
    # vortex contribution on a panel from others
    for i, panel_i in enumerate(panels):
        for j, panel_j in enumerate(panels):
            if i != j:
                A[i, j] = -0.5 / np.pi * integral(panel_i.xc, panel_i.yc,
                                                  panel_j,
                                                  panel_i.sin_beta,
                                                  -panel_i.cos_beta)
    return A


def kutta_condition(A_source, B_vortex):
    """
    Builds the Kutta condition array.

    Parameters
    ----------
    A_source: 2D Numpy array of floats
        Source contribution matrix for the normal velocity.
    B_vortex: 2D Numpy array of floats
        Vortex contribution matrix for the normal velocity.

    Returns
    -------
    b: 1D Numpy array of floats
        The left-hand side of the Kutta-condition equation.
    """
    b = np.empty(A_source.shape[0] + 1, dtype=float)
    # matrix of source contribution on tangential velocity
    # is the same than
    # matrix of vortex contribution on normal velocity
    b[:-1] = B_vortex[0, :] + B_vortex[-1, :]
    # matrix of vortex contribution on tangential velocity
    # is the opposite of
    # matrix of source contribution on normal velocity
    b[-1] = - np.sum(A_source[0, :] + A_source[-1, :])
    return b


def build_singularity_matrix(A_source, B_vortex):
    """
    Builds the left-hand side matrix of the system
    arising from source and vortex contributions.

    Parameters
    ----------
    A_source: 2D Numpy array of floats
        Source contribution matrix for the normal velocity.
    B_vortex: 2D Numpy array of floats
        Vortex contribution matrix for the normal velocity.

    Returns
    -------
    A:  2D Numpy array of floats
        Matrix of the linear system.
    """
    A = np.empty((A_source.shape[0] + 1, A_source.shape[1] + 1), dtype=float)
    # source contribution matrix
    A[:-1, :-1] = A_source
    # vortex contribution array
    A[:-1, -1] = np.sum(B_vortex, axis=1)
    # Kutta condition array
    A[-1, :] = kutta_condition(A_source, B_vortex)
    return A


def build_freestream_rhs(panels, freestream):
    """
    Builds the right-hand side of the system
    arising from the freestream contribution.

    Parameters
    ----------
    panels: 1D array of Panel objects
        List of panels.
    freestream: Freestream object
        Freestream conditions.

    Returns
    -------
    b: 1D Numpy array of floats
        Freestream contribution on each panel and on the Kutta condition.
    """
    b = np.empty(panels.size + 1, dtype=float)
    # freestream contribution on each panel
    for i, panel in enumerate(panels):
        b[i] = -freestream.u_inf * np.cos(freestream.alpha - panel.beta)
    # freestream contribution on the Kutta condition
    b[-1] = -freestream.u_inf * (np.sin(freestream.alpha - panels[0].beta) +
                                 np.sin(freestream.alpha - panels[-1].beta))
    return b


def compute_tangential_velocity(panels, freestream, gamma, A_source, B_vortex):
    """
    Computes the tangential surface velocity.

    Parameters
    ----------
    panels: 1D array of Panel objects
        List of panels.
    freestream: Freestream object
        Freestream conditions.
    gamma: float
        Circulation density.
    A_source: 2D Numpy array of floats
        Source contribution matrix for the normal velocity.
    B_vortex: 2D Numpy array of floats
        Vortex contribution matrix for the normal velocity.
    """
    A = np.empty((panels.size, panels.size + 1), dtype=float)
    # matrix of source contribution on tangential velocity
    # is the same than
    # matrix of vortex contribution on normal velocity
    A[:, :-1] = B_vortex
    # matrix of vortex contribution on tangential velocity
    # is the opposite of
    # matrix of source contribution on normal velocity
    A[:, -1] = -np.sum(A_source, axis=1)
    # freestream contribution
    b = freestream.u_inf * np.sin([freestream.alpha - panel.beta
                                   for panel in panels])

    strengths = np.append([panel.sigma for panel in panels], gamma)

    tangential_velocities = np.dot(A, strengths) + b

    for i, panel in enumerate(panels):
        panel.vt = tangential_velocities[i]


def compute_pressure_coefficient(panels, freestream):
    """
    Computes the surface pressure coefficients.

    Parameters
    ----------
    panels: 1D array of Panel objects
        List of panels.
    freestream: Freestream object
        Freestream conditions.
    """
    for panel in panels:
        panel.cp = 1.0 - (panel.vt / freestream.u_inf) ** 2


class PanelSolution(object):
    """
    Given an aerofoil object, angle of attack, and number of panels to
    discretise the geometry, compute pressure distribution data.
    """

    def __init__(self, aerofoil, alpha: Hint.num, N: int = None):
        # Recast as necessary
        alpha = float(alpha)
        N = 100 if N is None else N

        # discretize geoemetry into panels
        self.panels = define_panels(aerofoil, N=N)

        # Compute source and vortex influence matrices
        A_source = source_contribution_normal(self.panels)
        B_vortex = vortex_contribution_normal(self.panels)

        # define freestream conditions
        self.freestream = Freestream(u_inf=1.0, alpha=alpha)

        A = build_singularity_matrix(A_source, B_vortex)
        b = build_freestream_rhs(self.panels, self.freestream)

        # solve for singularity strengths
        strengths = np.linalg.solve(A, b)

        # store source strength on each panel
        for i, panel in enumerate(self.panels):
            panel.sigma = strengths[i]

        # store circulation density
        gamma = strengths[-1]

        # tangential velocity at each panel center.
        compute_tangential_velocity(
            self.panels, self.freestream, gamma, A_source, B_vortex)

        # surface pressure coefficient
        compute_pressure_coefficient(self.panels, self.freestream)

        # !!! Without wake panels, Cl and Cd cannot be trusted apparently
        # https://www.symscape.com/blog/why_use_panel_method
        # Compute normal and tangent force coefficients, then inviscid cl and cd
        # cn = sum([-pnl.cp * pnl.length * pnl.sin_beta for pnl in self.panels])
        # ct = sum([-pnl.cp * pnl.length * pnl.cos_beta for pnl in self.panels])
        # cl = cn * self.freestream.cos_alpha - ct * self.freestream.sin_alpha
        # cd = cn * self.freestream.sin_alpha + ct * self.freestream.cos_alpha

        return

    def show_Cp(self):
        """Plot surface pressure coefficient."""
        from matplotlib import pyplot as plt

        fig, ax = plt.subplots(1, dpi=140)

        ax.grid()
        ax.set_xlabel('$x/c$', fontsize=16)
        ax.set_ylabel('$C_p$', fontsize=16)
        ax.plot([panel.xc for panel in self.panels if panel.loc == 'upper'],
                [panel.cp for panel in self.panels if panel.loc == 'upper'],
                label='upper surface',
                color='r', linestyle='-', linewidth=2, marker='o',
                markersize=6)
        ax.plot([panel.xc for panel in self.panels if panel.loc == 'lower'],
                [panel.cp for panel in self.panels if panel.loc == 'lower'],
                label='lower surface',
                color='b', linestyle='-', linewidth=1, marker='o',
                markersize=6)

        ax.legend(loc='best', prop={'size': 16})
        ax.set_xlim(-0.1, 1.1)
        ax.set_ylim(None, 1.5)
        ax.set_ylim(*ax.get_ylim()[::-1])
        ax.set_title(f'Number of panels: {self.panels.size}', fontsize=16)
        plt.show()
        return


class InviscidPanelMethod(PanelSolution, AerofoilSolution):
    """A method for deriving the pressure distribution on an aerofoil."""

    def __init__(self, aerofoil, alpha: Hint.num, N: int = None):
        """
        Args:
            aerofoil: Aerofoil object.
            alpha: Angle of attack.
            N: Number of discretised points in the aerofoil's surface.
        """
        # Make super class call
        super().__init__(aerofoil, alpha, N)

        # !!! Without wake panels, Cl and Cd cannot be trusted apparently
        # https://www.symscape.com/blog/why_use_panel_method
        # Compute normal and tangent force coefficients, then inviscid cl and cd
        # cn = sum([-pnl.cp * pnl.length * pnl.sin_beta for pnl in self.panels])
        # ct = sum([-pnl.cp * pnl.length * pnl.cos_beta for pnl in self.panels])
        # cl = cn * self.freestream.cos_alpha - ct * self.freestream.sin_alpha
        # cd = cn * self.freestream.sin_alpha + ct * self.freestream.cos_alpha

        return



"""Methods relating to aerofoil profile generation and performance."""
from functools import cache
import re

import numpy as np
import requests
from scipy.optimize import minimize
from sectionproperties.analysis.section import Section
from sectionproperties.pre.geometry import Geometry

from carpy.utility import Hint, cast2numpy, isNone, point_curvature, point_diff
from ._generators import (
    NACA4DigitSeries, NACA4DigitModifiedSeries,
    NACA5DigitSeries, NACA5DigitModifiedSeries,
    NACA16Series
)

__all__ = ["NewAerofoil", "Aerofoil"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Support functions
# ---------------------------------------------------------------------------- #


def parse_datfile(coordinates) -> tuple[np.ndarray, np.ndarray]:
    """
    Given information to describe an aerofoil's profile, parse it into two
    arrays (describing upper and lower surface points, respectively). The
    data must be two-dimensional, and contain a point passing through (0, 0).

    Args:
        coordinates: Raw Selig format, raw Lednicer format, tuple of arrays,
            list of arrays, or an array describing aerofoil geometry. The data
            must be two dimensional, with one axis of size two.

    Returns:
        tuple: Two 2D arrays, describing points in upper and lower surfaces of
            the geometry.

    """
    # If instantiated with a string
    if isinstance(coordinates, str):
        parsed_groups = [[]]
        for line in coordinates.splitlines():
            # If line contains any A-z characters, invalidate that line
            if not isNone(re.match(r"[A-z]", line)):
                parsed_groups.append([])
            # If line contains 2 numbers, np.array contents and add to group
            elif len(matches := re.findall(r"[-+.e\d]+", line)) == 2:
                parsed_groups[-1].append(np.array(matches, dtype=float))
            # If line is empty, prepare next grouping
            elif line == "":
                parsed_groups.append([])
        else:
            # No aerofoil is described with 1 coordinate
            coordinates = [np.array(x) for x in parsed_groups if len(x) > 1]

    # If instantiated with an array
    elif isinstance(coordinates, np.ndarray):
        coordinates = [coordinates]

    # If instantiated with a tuple or list of arrays
    elif isinstance(coordinates, (tuple, list)) and all(
            [isinstance(x, np.ndarray) for x in coordinates]):
        pass

    else:
        errormsg = f"Unsupported input type: {type(coordinates)}"
        raise NotImplementedError(errormsg)

    # Sanity check: Coordinates should be 2D with array shape (n, 2)
    for i, array in enumerate(coordinates):
        if (dims := array.ndim) != 2:
            raise ValueError(f"Coordinate array should be 2D (got {dims})")
        if array.shape[0] == 2:
            coordinates[i] = array.T

    # All non-dimensional coordinate descriptions pass through (0, 0)
    # if Selig style (continuous surface)
    if (n_arrays := len(coordinates)) == 1:
        zeroes_idx = np.where(~coordinates[0].any(axis=1))[0][0]
        coordinates = \
            [coordinates[0][:zeroes_idx + 1][::-1], coordinates[0][zeroes_idx:]]
    # if Lednicer style
    elif n_arrays == 2:
        pass
    else:
        errormsg = (
            f"Got {n_arrays} arrays that could describe coordinates when "
            f"only 1 or 2 arrays are expected"
        )
        raise ValueError(errormsg)

    # The upper surface's ordinates average greater than lower surface's...
    surface_u, surface_l = coordinates
    if np.mean(surface_u[:, 1]) < np.mean(surface_l[:, 1]):
        surface_u, surface_l = surface_l, surface_u  # ..., swap if they weren't

    return surface_u, surface_l


def cast2points(array: Hint.nums):
    """
    Casts an array of 2d coordinates (or points) into points.

    Args:
        array: An array of 2d points and coordinates.

    Returns:
        An array of 2d points of shape (n, 2).

    """
    # Recast as necessary
    array = cast2numpy(array)

    assert array.ndim == 2, f"expected 2d array, got {array.ndim=}"

    if array.shape[1] == 2:
        pass
    elif array.shape[0] == 2:
        array = array.T
    else:
        raise ValueError(f"Expected array of shape (n, 2), got {array.shape}")

    return array


def find_camber(points: np.ndarray, idx_le: int, dxtarget=None) -> np.ndarray:
    """
    Find a set of points that describe the camber line of an aerofoil.

    Args:
        points: Points arranged CCW from TE to LE to TE, describing an aerofoil.
        idx_le: The index of a point in points to be considered the LE.
        dxtarget: A target spacing between camber line designation points.

    Returns:
        An array of points that describe the camber line.

    """
    # Recast as necessary
    dxtarget = 0.075 if dxtarget is None else dxtarget

    # Identify points on the aerofoil to perform camber analysis from
    # ... it's fine to use linear (not cosine!) spacing as the curvature is weak
    upper_points = points[:idx_le + 1]
    camber_xtarget = np.linspace(0, 1, int(1 / dxtarget))
    js = []
    for i in range(1, len(camber_xtarget) - 1):
        # Search upper surface for bounding points, and map camber to it
        for j in range(0, len(upper_points) - 1)[::-1]:
            if 0 < camber_xtarget[i] <= upper_points[j, 0] < 1:
                js.append(j)
                break
    else:
        js = sorted(set(js))  # Indices to use for camber locating

    # Compute normal vectors of points we wish to stage camber circles from
    camber_xy0 = upper_points[js]
    lower_points = points[idx_le:]
    norms = np.ones((len(js), 2))  # Empty array
    norms[:, 1] = -1 / point_diff(*np.fliplr(camber_xy0).T)  # Slopes
    norms[:] = [norm / np.linalg.norm(norm) for norm in norms]  # Normalise
    norms[norms[:, 1] > 0] = -norms[norms[:, 1] > 0]  # Orientation of norm

    # Use the normal vectors to construct camber finding circles
    camber_points = []
    for i, source_point in enumerate(camber_xy0):
        radii = []
        for point in lower_points:
            def f_opt(r1):
                """Opt. func based on radius of inscribed camber circle."""
                camber_point = source_point + r1 * norms[i]
                r2 = np.sum((point - camber_point) ** 2) ** 0.5
                result = abs(r1 - r2)  # difference between radii
                return result

            radius = minimize(
                f_opt, np.zeros(1), bounds=((0, None),), tol=1e-4).x
            radii.append(float(radius))

        camber_points.append(source_point + min(radii) * norms[i])

    # Include the obvious points of leading and trailing edge (0, 0), (1, 0)
    camber_points.insert(0, np.array([1, 0]))
    camber_points.append(np.array([0, 0]))

    return np.array(camber_points[::-1])


# ============================================================================ #
# Private Aerofoil class
# ---------------------------------------------------------------------------- #

class ProceduralProfiles(object):
    """A collection of procedural aerofoil profile generators."""

    @staticmethod
    def NACA(code: str, N=100):
        """
        Create a NACA 4-digit, 4-digit modified, 5-digit, 5-digit modified,
        or 16 series aerofoil.

        Args:
            code: The aerofoil code.
            N: The number of control points on each of upper and lower surfaces.
                Optional, defaults to 100 points on each surface (N=100).

        Returns:
            An Aerofoil object.

        Examples:

            >>> naca_codes = {
            ...     "4-digit": "2412", "4-digit modified": "2412-63",
            ...     "5-digit": "24012", "5-digit modified": "24012-33",
            ...     "16": "16-012", "16 modified": "16-012,a=0.5"
            ... }


            >>> for _, (_, v) in enumerate(naca_codes.items()):
            >>>     ProceduralProfiles.NACA(code=v, N=60).show()

        """
        naca_classes = [
            NACA4DigitSeries, NACA4DigitModifiedSeries,
            NACA5DigitSeries, NACA5DigitModifiedSeries,
            NACA16Series
        ]
        for naca_class in naca_classes:
            # noinspection PyProtectedMember
            if re.match(naca_class._pattern_valid, code):
                aerofoil = Aerofoil(points=naca_class(code).nd_xy(N=N))
                return aerofoil
        else:
            errormsg = (
                f"{code=} is not a recognised member of any of the following "
                f"series: {[x.__name__ for x in naca_classes]}"
            )
            raise ValueError(errormsg)


# ============================================================================ #
# Public-facing Aerofoil class
# ---------------------------------------------------------------------------- #

class Aerofoil(object):
    """
    A class for modelling the non-dimensional geometry and performance of
    aerofoils.
    """

    def __init__(self, *, points: Hint.nums = None,
                 upper_points: Hint.nums = None,
                 lower_points: Hint.nums = None):
        """
        Keyword Args:
            points: A 2d array of points describing the aerofoil profile.
            upper_points: A 2d array of points describing the upper surface
                of the aerofoil.
            lower_points: A 2d array of points describing the upper surface
                of the aerofoil.

        Notes:
            Either the 'points' keyword argument is required to describe the
            aerofoil geometry, or the 'upper_points' and 'lower_points' in
            concert.

        """
        # Recast as necessary (we need to regularise coordinate arguments)
        arg0 = not isNone(points)
        arg12 = not all(isNone(upper_points, lower_points))

        # ... no points are given
        if arg0 is False and arg12 is False:
            errormsg = f"{type(self).__name__}.__init__ requires 'points' or " \
                       f"'upper_points' and 'lower_points' are provided."
            raise ValueError(errormsg)

        # ... geometry is overdefined by points
        elif arg0 is True and arg12 is True:
            errormsg = f"{type(self).__name__}.__init__ received too many " \
                       f"arguments describing the aerofoil's geometry."
            raise ValueError(errormsg)

        # ... __init__ received one array of point coordinates
        elif arg0:
            points = cast2points(points)
            avg_start = points[:(arraylen := len(points)) // 2, 1].mean()
            avg_end = points[arraylen // 2:, 1].mean()
            if avg_start < avg_end:  # points are probably clockwise, cast CCW
                points = points[::-1]

        # ... __init__ received two arrays of point coordinates
        else:
            upper_points, lower_points = \
                map(cast2points, [upper_points, lower_points])
            if np.diff(upper_points[:, 0]).mean() > 0:  # diff suggest clockwise
                upper_points = upper_points[::-1]
            if np.diff(lower_points[:, 0]).mean() < 0:  # diff suggest clockwise
                lower_points = lower_points[::-1]
            # If leading edge points are duplicated, omit before concatenation
            if np.equal(upper_points[-1], lower_points[0]).all():
                points = np.concatenate([upper_points[:-1], lower_points])
            else:
                points = np.concatenate([upper_points, lower_points])
            del upper_points, lower_points

        self._points = points  # This is a counter-clockwise sequence from TE
        self._section = None  # Empty, until its time to compute it via property

        # Determine the coordinates representing the leading edge from curvature
        self._curvature = abs(point_curvature(*points.T))  # Unsigned curvature

        return

    def __add__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(f"Cannot add {type(self)=} to {type(other)=}")

        # It is not necessarily the case that self and other's upper and lower
        # surfaces share abscissa. For this reason, each surface on each
        # aerofoil must be interpolated.
        idx_le_self = int(np.argmax(self._curvature))
        idx_le_other = int(np.argmax(other._curvature))

        # Determine the number of points which shall define a new surface
        n_points = max(
            len((upper_self := self._points[:idx_le_self + 1][::-1])),
            len((upper_other := other._points[:idx_le_other + 1][::-1])),
            len((lower_self := self._points[idx_le_self:])),
            len((lower_other := other._points[idx_le_other:]))
        )
        xnew = (np.cos(np.linspace(np.pi, 0, n_points)) + 1) / 2
        y_u = np.interp(xnew, *upper_self.T) + np.interp(xnew, *upper_other.T)
        y_l = np.interp(xnew, *lower_self.T) + np.interp(xnew, *lower_other.T)

        new_object = type(self)(
            upper_points=np.vstack([xnew, y_u]).T,
            lower_points=np.vstack([xnew, y_l]).T
        )
        return new_object

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        # Typechecking
        if not isinstance(other, Hint.nums.__args__):
            raise TypeError(f"Cannot multiply {type(self)=} by {type(other)=}")
        elif isinstance(other, Hint.num.__args__):
            multiplier = cast2numpy([1, other])
        else:
            multiplier = cast2numpy(other)
        # Simply transform the coordinates
        new_object = type(self)(points=self._points * multiplier)
        return new_object

    def __rmul__(self, other):
        return self.__mul__(other)

    @property
    def le_radius(self):
        """Radius of the leading edge."""
        return 1 / self._curvature.max()

    @property
    def perimeter(self):
        """(Non-dimensional) perimeter ratio to a reference chord of 1."""
        # Take dxs and dys, use pythagoras to find lengths between points, & sum
        return ((np.diff(self._points, axis=0) ** 2).sum(axis=1) ** 0.5).sum()

    def _camber_points(self, step_target: Hint.num,
                       fast: bool = None) -> np.ndarray:
        """
        An array of points that describe the aerofoil's camber line.

        Args:
            step_target: Approximate distance between camber designating points.
            fast: Boolean flag, True if user wishes to use a lower accuracy but
                faster approximation. Optional, defaults to True. See Notes for
                more details.

        Returns:
            A 2D array of points describing the aerofoil's camber line.

        Notes:
            In fast mode, the camber is determined by averaging the coordinates
                of upper and lower surfaces. In slow mode, camber points are
                determined by finding the parameters of circles inscribing the
                aerofoil geometry. This method is not robust, and can have
                trouble - especially with thinner geometry like that found at
                the trailing edge.

        """
        # Recast as necessary
        fast = True if fast is None else fast

        idx_le = int(np.argmax(self._curvature))  # Index of the leading edge

        if fast is True:
            xs = np.arange(0, 1 + step_target, step_target)
            y_u = np.interp(xs, *self._points[:idx_le + 1][::-1].T)
            y_l = np.interp(xs, *self._points[idx_le:].T)
            points = np.column_stack((xs, (y_u + y_l) / 2))
            points[0] = 0, 0
            points[-1] = 1, 0
        else:
            points = find_camber(
                points=self._points, idx_le=idx_le, dxtarget=step_target)

        return points

    def _thickness_function(self, orientation: Hint.num = None) -> Hint.func:
        # Recast as necessary
        convention = "American" if orientation is None else "British"

        if convention == "American":
            errormsg = (
                "Sorry, this is unavailable right now. You can still proceed "
                "with identifying aerofoil thickness through the British "
                "convention which is possible by passing the 'orientation' arg."
            )
            raise NotImplementedError(errormsg)
        else:
            # Cosine spaced sampling points
            xc = (np.cos(np.linspace(np.pi, 0, 100)) + 1) / 2

            # Translate the leading edge point to (0, 0) (if it isn't already??)
            idx_le = int(np.argmax(self._curvature))  # Index of leading edge
            points = self._points - self._points[idx_le]

            # Transform geometry into the desired orientation for thickness eval
            cos_o, sin_o = np.cos(orientation), np.sin(orientation)
            rot_matrix = np.array([[cos_o, -sin_o], [sin_o, cos_o]])
            points = np.array([rot_matrix @ point for point in points])

            # Compute thickness
            upper_points = points[:idx_le + 1][::-1]  # Order from left to right
            lower_points = points[idx_le:]
            t_hi = np.interp(xc * cos_o, *upper_points.T)
            t_lo = np.interp(xc * cos_o, *lower_points.T)

            def thickness(x: Hint.nums):
                """
                The available thickness of the aerofoil in an orientation
                traditionally perpendicular to the chord line.
                """
                # Recast as necessary
                x = cast2numpy(x)
                return np.interp(x, xc, t_hi - t_lo, left=np.nan, right=np.nan)

            thickness.__doc__ = (
                f"\n    The available thickness of the aerofoil in an"
                f"\n    orientation traditionally perpendicular to the chord"
                f"\n    line. In this case, the available thickness is measured"
                f"\n    in the vertical direction after the aerofoil geometry"
                f"\n    has been rotated counter-clockwise by {orientation:.3f}"
                f" radians."
                f"\n    Args:"
                f"\n    \tx: The chordwise position at which thickness should"
                f"\n    \t\tbe evaluated."
                f"\n"
                f"\n    Returns:"
                f"\n    \tThe thickness of the aerofoil.\n"
            )

        return thickness

    @property
    def section(self) -> Section:
        """Section properties, for geometric analysis of the aerofoil."""
        # If section has been generated previously, reuse it
        if self._section is not None:
            return self._section

        # Create a section...
        # Describe the connectivity of the points in a nested list
        facets = (
                         np.arange(len(self._points))[:, None] +
                         np.array([0, 1])
                 ) % len(self._points)
        facets = [list(x) for x in facets]
        # Identify the inside region of the aerofoil as 10% behind leading edge
        control_points = [[0.1, 0]]

        # Create a geometry object
        section_geometry = Geometry.from_points(
            points=self._points,
            facets=facets,
            control_points=control_points
        )
        section_geometry.create_mesh(mesh_sizes=[1])
        # Use the geometry object to instantiate a Section object
        self._section = Section(section_geometry)
        self._section.calculate_geometric_properties()  # Update geometric props
        return self._section

    def show(self):
        """Simple 2D render of the aerofoil geometry."""
        from matplotlib import pyplot as plt

        fig, ax = plt.subplots(1, dpi=140)
        ax.plot(*self._points.T, c="k")
        camber_points = self._camber_points(step_target=0.05, fast=True)
        ax.plot(*camber_points.T, c="orange")

        # Make the primary plot pretty
        fig.canvas.manager.set_window_title(f"{self}.show()")
        ax.set_title("Aerofoil 2D Profile")
        ax.set_aspect(1)
        ax.set_xlabel("x/c")
        ax.set_ylabel("y/c")
        ax.grid(alpha=0.3)

        plt.show()

        return


class NewAerofoil(object):
    """
    A class of methods for generating non-dimensional aerofoil geometries.
    """

    # From online sources
    @classmethod
    @cache
    def from_url(cls, url: str):
        """
        Return an aerofoil object, given the URL to a coordinates data file.

        Args:
            url: A web-URL to a Selig or Lednicer format coordinate data file.

        Returns:
            An Aerofoil object.

        Examples:

            >>> my_url = "https://m-selig.ae.illinois.edu/ads/coord/n0012.dat"
            >>> n0012 = NewAerofoil.from_url(url=my_url)
            >>> n0012.show()

        """
        response = requests.get(url=url)

        # On successful request
        if response.status_code == 200:
            upper_pts, lower_pts = parse_datfile(response.text)
            aerofoil = Aerofoil(upper_points=upper_pts, lower_points=lower_pts)
        else:
            raise ConnectionError("Couldn't access given URL")

        return aerofoil

    # From local filepath
    @classmethod
    def from_path(cls, filepath):
        """
        Return an aerofoil object, given the path to a coordinates data file.

        Args:
            filepath: A path to a Selig or Lednicer format coordinate data file.

        Returns:
            An Aerofoil object.

        Examples:

            >>> my_path = "C:\\Users\\Public\\Documents\\naca0012.txt"
            >>> n0012 = NewAerofoil.from_path(filepath=my_path)
            >>> n0012.show()

        """
        with open(filepath, "r") as f:
            filecontents = f.read()

        upper_pts, lower_pts = parse_datfile(filecontents)
        aerofoil = Aerofoil(upper_points=upper_pts, lower_points=lower_pts)

        return aerofoil

    # From procedural generators of aerofoil geometry
    from_method = ProceduralProfiles



"""A module of tools to make things look prettier."""
import re

__all__ = []
__author__ = "Yaseen Reza"


# ============================================================================ #
# Subscripting and superscripting
# ---------------------------------------------------------------------------- #

class Unicodify(object):
    """Methods based on sub- and super- scripting in Unicode strings."""

    @staticmethod
    def subscript_all(text: str) -> str:
        """Using unicode, attempt to subscript anything in the input."""
        unicodemapping = {
            # unicode prefix: dict(zip(character2translate, unicode suffix))
            "0x208": dict(zip("0123456789+-=()", "0123456789ABCDEF")),
            "0x209": dict(zip("aeoxəhklmnpst", "0123456789ABCDEF"))
        }
        # Create a map from 'character2translate' to its unicode translation
        mapping = str.maketrans({
            target: chr(int(f"{prefix}{suffix}", 16))
            for prefix, submap in unicodemapping.items()
            for target, suffix in submap.items()
        })
        return text.translate(mapping)

    @staticmethod
    def superscript_all(text: str) -> str:
        """Using unicode, attempt to superscript anything in the input."""
        unicodemapping = {
            # unicode prefix: dict(zip(character2translate, unicode suffix))
            "0x00B": dict(zip("231", "239")),
            "0x207": dict(zip("0i456789+-=()n", "01456789ABCDEF")),
        }
        # Create a map from 'character2translate' to its unicode translation
        mapping = str.maketrans({
            target: chr(int(f"{prefix}{suffix}", 16))
            for prefix, submap in unicodemapping.items()
            for target, suffix in submap.items()
        })
        return text.translate(mapping)

    @classmethod
    def mathscript_safe(cls, text: str, /) -> str:
        """
        Replaces, when possible, text with unicode sub and/or superscript
        variants.

        Args:
            text: The string of text to apply the unicode transformation to.

        Returns:
            A string where superscript and subscript substitutions have been
            made.

        Examples: ::

            >>> Unicodify.mathscript_safe("g_{0} = 9.81 ms^{2}")
            # Outputs: g₀ = 9.81 ms²

        """
        # Identify text that should be made super or subscript
        re_numscript = re.compile(r"[\^_]{[^(?!{})]+?}")
        text2translate = set(re_numscript.findall(text))

        # Apply super or subscript translation
        for numscript in text2translate:
            translator = {
                "^": cls.superscript_all, "_": cls.subscript_all
            }[numscript[0]]
            # Use indexing [2:-1] to ignore ^{} and _{} patterns
            text = text.replace(numscript, translator(numscript[2:-1]))

        return text

    @classmethod
    def chemical_formula(cls, text: str, /) -> str:
        """
        Formats molecular formula with subscripted numbers of atoms.

        Args:
            text: The string of text on which to apply the Unicode transform.

        Returns:
            Chemical formula with subscripted numbers of atoms.

        Examples: ::

            >>> print(Unicodify.chemical_formula("2(C2H5)2O"))
            '2(C₂H₅)₂O'

        """
        prettystring = "".join([
            Unicodify.subscript_all(char)
            if (char.isnumeric() and i != 0) else char
            for i, char in enumerate(re.findall(r"\d+|[A-z()]+", text))
        ])
        return prettystring

    @classmethod
    def superscript_trailingnums(cls, text: str, /) -> str:
        """
        Formats trailing numerical characters in a word with superscripting.

        Args:
            text: The string of text on which to apply the Unicode transform.

        Returns:
            String with the trailing numbers superscripted.

        Examples: ::

            >>> print("Neon has the electron arrangement", end=" ")
            >>> print(Unicodify.superscript_trailingnums("1s2 2s2 2p6"))
            'Neon has the electron arrangement 1s² 2s² 2p⁶'

        """

        def match_handler(matchobj):
            """Take a match object and superscript all the text contained."""
            return cls.superscript_all(matchobj.group(0))

        return re.sub(r"\d+\s|\d+$", match_handler, text)


__all__ += [Unicodify.__name__]


# ============================================================================ #

class Pretty(object):
    """Assorted methods for making pretty outputs."""

    @staticmethod
    def temperature(T, unit: str = None):
        """
        Make a pretty temperature string given the temperature and unit.

        Args:
            T: Scalar quantity of temperature that can be cast as a float.
            unit: One of "degC", "degF", "degR", or "K" units of temperature.

        Returns:
            A string for neatly displaying the temperature.

        """
        # Prefix (?)
        prefix = "+" if T >= 0 else ""

        # Temperature string
        temperature = f"{float(T)}".rstrip("0") if float(T) != 0 else "0"
        temperature = temperature.rstrip(".")

        # Suffix
        unit = "degC" if unit is None else unit
        suffix = {
            "degC": u"\N{DEGREE SIGN}C",
            "degF": u"\N{DEGREE SIGN}F",
            "degR": u"\N{DEGREE SIGN}R",
            "K": " K"
        }[unit]

        temperature_string = f"{prefix}{temperature}{suffix}"

        return temperature_string

    @staticmethod
    def signed(text, /):
        """
        Given a string representing a number, add a plus-sign to the front of
        the string if it represents a value above 0.

        Args:
            text: A string of text representing the number to make pretty.

        Returns:
            A signed string of text.

        """
        # Hopefully try-except architecture this is input-type agnostic
        try:
            number = float(text)
        except ValueError:
            pass
        else:
            if number > 0:
                text = f"+{text}"
        finally:
            return f"{text}"


__all__ += [Pretty.__name__]



"""Methods relating to aerofoil profile generation."""
import re

import numpy as np

from carpy.utility import Hint, cast2numpy

__author__ = "Yaseen Reza"


# ============================================================================ #
# Base classes for procedural aerofoil generators
# ---------------------------------------------------------------------------- #
class BaseProfile(object):
    """General parameters of aerofoil profiles."""

    _f_nd_yz: Hint.func = lambda self, x: x * 0
    _f_nd_yt: Hint.func = lambda self, x: x * 0

    def nd_yz(self, x: Hint.nums) -> np.ndarray:
        """
        Non-dimensional camber line (aerofoil mean-line) ordinate.
        """
        # Recast as necessary
        x = cast2numpy(x)
        return self._f_nd_yz(x)

    def nd_yt(self, x: Hint.nums):
        """
        Non-dimensional thickness distribution, as per American convention.
        """
        # Recast as necessary
        x = cast2numpy(x)
        return self._f_nd_yt(x)

    def nd_xy(self, N: int, *, concatenate: bool = None, closeTE: bool = None):
        """
        Non-dimensional coordinates for each of upper and lower surfaces.
        """
        # Recast as necessary
        concatenate = True if concatenate is None else concatenate
        closeTE = False if closeTE is None else closeTE

        # Cosine distribution for higher resolution near the leading edge
        dphi = (np.pi / 2) * (1 / (N - 1))
        xs = 1 - np.cos(np.arange(0, N) * dphi)

        # Gradient of camber line
        eps = 1e-6
        xs_addeps = np.clip(xs + eps, None, 1)
        xs_subeps = np.clip(xs - eps, 0, None)
        dzdx = (self.nd_yz(xs_addeps) - self.nd_yz(xs_subeps)) / (2 * eps)

        # Ordinate rotation angle
        theta = np.arctan(dzdx)

        # Compute upper and lower coordinates
        zs = self.nd_yz(xs)  # Camber line
        ts = self.nd_yt(xs)  # Thickness distribution
        u_abscissa = xs - ts * np.sin(theta)
        u_ordinate = zs + ts * np.cos(theta)
        l_abscissa = xs + ts * np.sin(theta)
        l_ordinate = zs - ts * np.cos(theta)

        # Package nicely and return
        u_coords = np.array([u_abscissa, u_ordinate])
        l_coords = np.array([l_abscissa, l_ordinate])

        if closeTE is True:
            te_coord = np.array([[1], [0]])
            u_coords = np.concatenate([u_coords, te_coord], axis=1)
            l_coords = np.concatenate([l_coords, te_coord], axis=1)

        if concatenate is True:
            # Force CCW traversal of upper surface and omit duplicated LE coord
            u_coords = np.flip(u_coords, axis=1)[:, :-1]
            return np.concatenate([u_coords, l_coords], axis=1)

        return u_coords, l_coords


class BaseNACA(BaseProfile):
    """General methods for NACA aerofoil objects in carpy."""

    _pattern_valid: str = NotImplemented

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Check that the code is valid
        if not self._is_valid(code=code):
            errormsg = f"{type(self).__name__} got an invalid NACA {code=}"
            raise ValueError(errormsg)

        # Create a private object for recording any parsed terms.
        self._parsed = type(
            "Parsed",
            (),
            {
                "LEri": None,  # Leading Edge radius index number
                "c_max_x": None,  # Chordwise location of maximum camber
                "c_max_y": None,  # Maximal camber ordinate
                "cli": None,  # Design section lift coefficient
                "cli_d": None,  # Range of drag bucket above and below cli
                "flag_experimental": None,  # Experimental aerofoils designation
                "flag_reflex": None,  # Boolean, reflexed camber or not
                "meanline_a": None,  # Extent of uniform loading
                "meanline_b": None,  # Extent of non-zero loading
                "meanline_cli": None,  # Design coefficient of lift for meanline
                "p_min_x": None,  # Chordwise location of minimum pressure
                "t_max": None,  # Maximum thickness to chord ratio
                "t_max_x": None,  # Chordwise location of maximum thickness
                "t0_max": None,  # Max thickness of original distribution
            }
        )()

        return

    @classmethod
    def _is_valid(cls, code: str):
        """Verify that an aerofoil code is a valid member of this class."""
        if re.match(cls._pattern_valid, code):
            return True
        return False

    @staticmethod
    def _pop_pattern(string: str, pattern, *args, **kwargs) -> tuple[str, str]:
        """
        Use regular expression matching to find and pop start of a string.

        Args:
            string: The string to be searched.
            pattern: A pattern expected to be found at the start of the string.
            *args: Miscellaneous arguments to pass to re.match().
            **kwargs: Miscellaneous keyword arguments to pass to re.match().

        Returns:
            A tuple containing the original string stripped of any matches, and
                the string matching the pattern specified by the user.

        """
        match = re.match(pattern, string, *args, **kwargs)

        if match is None:
            raise ValueError(f"Start of '{string}' didn't match {pattern=}")

        _, endofmatch = match.span()
        string_sliced = match.string[endofmatch:]  # String with prefix removed
        string_popped = match.group()  # String that matched the desired pattern

        return string_sliced, string_popped


# ============================================================================ #
# Procedural aerofoil generators
# ---------------------------------------------------------------------------- #

class NACA4DigitSeries(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data
        -   NASA Technical Memorandum 4741

    """
    _pattern_valid = r"^([0]{2}|[1-9]{2})\d{2}$"

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse
        code, c_max_y = self._pop_pattern(code, r"\d")
        code, c_max_x = self._pop_pattern(code, r"\d")
        _, t_max = self._pop_pattern(code, r"\d{2}")

        # Recast
        self._parsed.c_max_y = float(c_max_y) / 100
        self._parsed.c_max_x = float(c_max_x) / 10
        self._parsed.t_max = float(t_max) / 100
        return

    def _f_nd_yz(self, x):
        """Private camber line function."""
        # Recast as necessary
        parsed = self._parsed

        # For symmetric aerofoils, this part is easy - camberline is just zero
        if parsed.c_max_y == 0 and parsed.c_max_x == 0:
            return x * 0

        # For cambered aerofoils...
        yz = np.where(
            x <= parsed.c_max_x,
            # Condition True
            parsed.c_max_y * (
                    (2 * parsed.c_max_x - x) * x) / parsed.c_max_x ** 2,
            # Condition False
            parsed.c_max_y * (
                    (1 - 2 * parsed.c_max_x) + 2 * parsed.c_max_x * x - x ** 2)
            / (1 - parsed.c_max_x) ** 2
        )
        return yz

    def _f_nd_yt(self, x):
        """Private thickness distribution function."""
        # Recast as necessary
        parsed = self._parsed

        # Constants of t/c = 0.20 polynomial for thickness
        a = np.array([0.2969, -0.1260, -0.3516, 0.2843, -0.1015])
        exp = np.array([0.5, 1.0, 2.0, 3.0, 4.0])

        # Ordinates for a xx% thickness aerofoil (by normalising t/c=0.20 form)
        def transform(xi):
            """Compute polynomial for thickness distribution."""
            return np.sum(a * xi ** exp) * (parsed.t_max / 0.20)

        yt = np.vectorize(transform)(x)
        return yt


class NACA4DigitModifiedSeries(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data
        -   NASA Technical Memorandum 4741

    """
    _pattern_valid = r"^([0]{2}|[1-9]{2})\d{2}-\d[2-6]$"

    # noinspection PyProtectedMember
    _f_nd_yz = NACA4DigitSeries._f_nd_yz  # Identical camber function

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse
        code, c_max_y = self._pop_pattern(code, r"\d")
        code, c_max_x = self._pop_pattern(code, r"\d")
        code, t_max = self._pop_pattern(code, r"\d{2}")
        code, _ = self._pop_pattern(code, r"-")
        code, LEri = self._pop_pattern(code, r"\d")
        _, t_max_x = self._pop_pattern(code, r"\d")

        # Recast
        self._parsed.c_max_y = float(c_max_y) / 100
        self._parsed.c_max_x = float(c_max_x) / 10
        self._parsed.t_max = float(t_max) / 100
        self._parsed.LEri = int(LEri)
        self._parsed.t_max_x = float(t_max_x) / 10
        return

    def _f_nd_yt(self, x):
        """Private thickness distribution function."""
        # Recast as necessary
        parsed = self._parsed
        t_max_x = parsed.t_max_x

        # Solve linear system to discover missing constants
        # Initialise constants a_i
        a = np.zeros(4)
        exponent = np.array([0.5, 1.0, 2.0, 3.0])
        # Modify a0 as a function of the leading edge radius index
        a0 = 0.2969
        if parsed.LEri != 9:
            a0 *= parsed.LEri / 6
        else:
            a0 *= 3 ** 0.5
        lhs = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [t_max_x ** 0.5, t_max_x, t_max_x ** 2, t_max_x ** 3],
            [0.5 * t_max_x ** -0.5, 1, 2 * t_max_x, 3 * t_max_x ** 2]
        ])
        rhs = np.array([a0, 0, 0.1, 0])
        a = np.linalg.solve(lhs, rhs)

        # Solve linear system to discover missing constants
        # Initialise constants d_i
        d0 = 0.002
        d1 = {
            0.2: 0.200, 0.3: 0.234, 0.4: 0.315, 0.5: 0.465, 0.6: 0.700
        }[t_max_x]
        lhs = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [1, (1 - t_max_x), (1 - t_max_x) ** 2, (1 - t_max_x) ** 3],
            [0, 1, 2 * (1 - t_max_x), 3 * (1 - t_max_x) ** 2],
        ])
        rhs = np.array([d0, d1, 0.1, 0])
        d = np.linalg.solve(lhs, rhs)

        # Ordinates for a xx% thickness aerofoil (by normalising t/c=0.20 form)
        scalar = (parsed.t_max / 0.20)

        def f_tc_half(xc):
            """Switch thickness distribution being used based on xc position."""
            if xc <= t_max_x:
                result = scalar * np.sum(a * xc ** exponent)
            else:
                result = scalar * np.sum(d * (1 - xc) ** np.arange(4))
            return result

        yt = np.vectorize(f_tc_half)(x)
        return yt


class NACA5DigitSeries(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data
        -   NASA Technical Memorandum 4741

    """
    _pattern_valid = r"^\d[2-5][01]\d{2}$"

    # noinspection PyProtectedMember
    _f_nd_yt = NACA4DigitSeries._f_nd_yt  # Identical thickness function

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse
        code, cli = self._pop_pattern(code, r"\d")
        code, c_max_x = self._pop_pattern(code, r"\d")
        code, flag_reflex = self._pop_pattern(code, r"\d")
        _, t_max = self._pop_pattern(code, r"\d{2}")

        # Recast
        self._parsed.cli = float(cli) * 3 / 20
        self._parsed.c_max_x = float(c_max_x) / 20
        self._parsed.flag_reflex = bool(int(flag_reflex))
        self._parsed.t_max = float(t_max) / 100
        return

    def _f_nd_yz(self, x):
        """Private camber line function."""
        # Recast as necessary
        parsed = self._parsed

        # Check for reflex, else continue as normal
        if parsed.flag_reflex is False:

            # Use look-up to find values of r, k1 that correspond with c_max_x
            i = [0.05, 0.10, 0.15, 0.20, 0.25].index(parsed.c_max_x)
            r = [0.0580, 0.1260, 0.2025, 0.2900, 0.3910][i]
            k1 = [361.400, 51.640, 15.957, 6.643, 3.230][i] * (parsed.cli / 0.3)

            # Camber function
            def f_zc(xc):
                """Camber line with chord position."""
                zc = (k1 / 6) * np.where(
                    xc <= r,
                    xc ** 3 - 3 * r * xc ** 2 + r ** 2 * (3 - r) * xc,
                    r ** 3 * (1 - xc)
                )
                return zc

            yz = np.vectorize(f_zc)(x)

        else:
            # Every single source I've read on this will provide the formulas
            # to generate the constants below (as well as precomputing them) but
            # they never match what I'm able to reproduce by implementing them

            # Use look-up to find  r, k1, n=(k2/k1) that correspond with x_cmax
            i = [0.10, 0.15, 0.20, 0.25].index(parsed.c_max_x)
            r = [0.1300, 0.2170, 0.3180, 0.4410][i]
            k1 = [51.990, 15.793, 6.520, 3.191][i] * (parsed.cli / 0.30)
            n = [0.000764, 0.00677, 0.0303, 0.1455][i]

            # Camber function
            def f_zc(xc):
                """Camber line with chord position."""
                zc = (k1 / 6) * np.where(
                    xc <= r,
                    # Condition True
                    (xc - r) ** 3
                    - n * (1 - r) ** 3 * xc - r ** 3 * xc + r ** 3,
                    # Condition False
                    n * (xc - r) ** 3
                    - n * (1 - r) ** 3 * xc - r ** 3 * xc + r ** 3
                )
                return zc

            yz = np.vectorize(f_zc)(x)

        return yz


class NACA5DigitModifiedSeries(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data
        -   NASA Technical Memorandum 4741

    """
    _pattern_valid = r"^\d[2-5][01]\d{2}-\d[2-6]$"

    # noinspection PyProtectedMember
    _f_nd_yz = NACA5DigitSeries._f_nd_yz  # Identical camber function
    # noinspection PyProtectedMember
    _f_nd_yt = NACA4DigitModifiedSeries._f_nd_yt  # Identical thickness function

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse
        code, cli = self._pop_pattern(code, r"\d")
        code, c_max_x = self._pop_pattern(code, r"\d")
        code, flag_reflex = self._pop_pattern(code, r"\d")
        code, t_max = self._pop_pattern(code, r"\d{2}")
        code, _ = self._pop_pattern(code, r"-")
        code, LEri = self._pop_pattern(code, r"\d")
        _, t_max_x = self._pop_pattern(code, r"\d")

        # Recast
        self._parsed.cli = float(cli) * 3 / 20
        self._parsed.c_max_x = float(c_max_x) / 20
        self._parsed.flag_reflex = bool(int(flag_reflex))
        self._parsed.t_max = float(t_max) / 100
        self._parsed.LEri = int(LEri)
        self._parsed.t_max_x = float(t_max_x) / 10
        return


class NACA6Series(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data

    """
    _pattern_valid = (
            r"^6\d"  # First two digits
            + r"((,\d)|(\(\d{3}\))|(_{\d})|(_{\(\d{2,3}\)}))?x?-"  # Upto hyphen
            + r"((\d)|(\(\d\.\d+\)))"  # Design lift coefficient
            + r"((\d{2})|(\(\d{2}\.\d+\)))"  # Relative thickness
            + r"(" + (
                "|".join(
                    [  # Suffix
                        "",  # Default mean-line (a = 1.0)
                        r"(,a=0?(\.\d+)?)|(,a=1(\.0+)?)",  # Mean-line only
                        r"(\{(((a=0?(\.\d+)?)|(a=1(\.0+)?)),"  # Mean-line(s)...
                        r"cli=-?\d?(\.\d+).?)+\})"
                        # ...and design lift coefficient(s)
                    ]
                )
            ) + ")$"
    )

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse: Series and minimum pressure position
        code, _ = self._pop_pattern(code, r"\d")
        code, p_min_x = self._pop_pattern(code, r"\d")

        # Parse terms before the hyphen
        cli_d = np.nan
        t0_max = None
        if re.match(r"^,\d", code):
            code, _ = self._pop_pattern(code, ",")
            code, cli_d = self._pop_pattern(code, r"\d")
        elif re.match(r"^_{\d}", code):
            code, _ = self._pop_pattern(code, r"_\{")
            code, cli_d = self._pop_pattern(code, r"\d")
            code, _ = self._pop_pattern(code, r"\}")
        elif re.match(r"^\(\d{3}\)", code):
            code, _ = self._pop_pattern(code, r"\(")
            code, cli_d = self._pop_pattern(code, r"\d")
            code, t0_max = self._pop_pattern(code, r"\d{2}")
            code, _ = self._pop_pattern(code, r"\)")
        elif re.match(r"^_{\(\d{3}\)}", code):
            code, _ = self._pop_pattern(code, r"_{\(")
            code, cli_d = self._pop_pattern(code, r"\d")
            code, t0_max = self._pop_pattern(code, r"\d{2}")
            code, _ = self._pop_pattern(code, r"\)}")
        elif re.match(r"^_{\(\d{2}\)}", code):
            code, _ = self._pop_pattern(code, r"_{\(")
            code, t0_max = self._pop_pattern(code, r"\d{2}")
            code, _ = self._pop_pattern(code, r"\)}")

        # Parse the hyphen (and ignore experimental tag)
        code, flag_experimental = self._pop_pattern(code, "^x?")
        code, _ = self._pop_pattern(code, "^-")

        # Parse design lift coefficient and maximum thickness
        if re.match(r"^\d", code):
            code, cli = self._pop_pattern(code, r"\d")
        else:
            code, _ = self._pop_pattern(code, r"\(|")
            code, cli = self._pop_pattern(code, r"\d(\.\d+)?")
            code, _ = self._pop_pattern(code, r"\)|")
        if re.match(r"^\d{2}", code):
            code, t_max = self._pop_pattern(code, r"\d{2}")
        else:
            code, _ = self._pop_pattern(code, r"\(|")
            code, t_max = self._pop_pattern(code, r"\d{2}(\.\d+)?")
            code, _ = self._pop_pattern(code, r"\)|")

        # Parse addenda
        meanline_cli = [float(cli) / 10]
        if re.match("^$", code):
            meanline_a = [1]
        elif re.match(r"^,a=\d?(\.\d+)?$", code):
            code, _ = self._pop_pattern(code, ",a=")
            meanline_a = [code]
        else:
            meanline_a = re.findall(r"a=(\d?(\.\d+)?)", code)
            meanline_cli = re.findall(r"cli=(-?\d?(\.\d+)?)", code)

        # Recast
        self._parsed.p_min_x = float(p_min_x) / 10
        self._parsed.cli_d = float(cli_d) / 10
        self._parsed.t0_max = float(t_max if t0_max is None else t0_max) / 100
        self._parsed.flag_experimental = bool(flag_experimental)
        self._parsed.cli = float(cli) / 10
        self._parsed.t_max = float(t_max) / 100
        self._parsed.meanline_a = np.array(meanline_a, dtype=np.float64)
        self._parsed.meanline_cli = np.array(meanline_cli, dtype=np.float64)
        return

    def _f_nd_yz(self, x):
        """
        Private camber line function.

        Notes:
            According to NACA report No.824, "For special purposes, load
            distributions other than those corresponding to the simple mean
            lines may be obtained by combining two or more types of mean line
            having positive or negative values of the design lift coefficient.
            The geometric and aerodynamic characteristics of such combinations
            may be obtained by algebraic addition of the values for the
            component mean lines". I interpret this as saying, if you have
            multiple mean line designations - the final mean line is simply the
            arithmetic sum of the component mean lines.
        """
        yz_cumulative = np.zeros_like(x)

        for i in range(len(self._parsed.meanline_a)):
            yz_cumulative += self._f_camber(
                cli=self._parsed.meanline_cli[i],
                a=self._parsed.meanline_a[i],
                b=1
            )(x)

        return yz_cumulative

    @classmethod
    def _f_camber(cls, cli: float, a: float = None, b: float = None):
        """Private function to create camberline associated with meanline."""
        # Recast as necessary
        a = a if a is not None else 1.0
        b = b if b is not None else 1.0

        # Unique case when there would otherwise be some crazy singularities
        if a == 1 and b == 1:
            def f_zc(xc):
                """Camber line function."""
                if xc == 0 or xc == 1:
                    return 0.0  # <-- Has to be a float or np.vectorize is sad!!
                zc = -cli / (4 * np.pi) * (
                        (1 - xc) * np.log(1 - xc) + xc * np.log(xc))
                return zc

        # The next case throws an error because it does not make physical sense.
        # If the aerofoil must maintain uniform loading, and then decrease to
        # zero loading by point b - it doesn't make sense for a to be >= b.
        elif a >= b:
            errormsg = (
                f"Extent of uniform loading '{a=}' cannot be greater than or "
                f"equal to the extent of non-zero loading '{b=}'"
            )
            raise ValueError(errormsg)

        # Every other valid case
        else:
            # Unique case when b == 1 would also cause singularities
            if b == 1:
                g = -1 / (1 - a) * (a ** 2 * (0.5 * np.log(a) - 0.25) + 0.25)
                h = 1 / (1 - a) * (
                        0.5 * (1 - a) ** 2 * np.log(1 - a) - 0.25 * (1 - a) ** 2
                ) + g
            else:
                g = -1 / (b - a) * (
                        a ** 2 * (0.5 * np.log(a) - 0.25)
                        - b ** 2 * (0.5 * np.log(b) - 0.25)
                )
                h = 1 / (b - a) * (
                        0.5 * (1 - a) ** 2 * np.log(1 - a)
                        - 0.5 * (1 - b) ** 2 * np.log(1 - b)
                        + 0.25 * (1 - b) ** 2
                        - 0.25 * (1 - a) ** 2
                ) + g

            def f_zc(xc):
                """Camber line function."""
                if xc == 0 or xc == 1:
                    return 0.0  # <-- Has to be float or np.vectorize is sad!!!
                zc = cli / (2 * np.pi * (a + b))
                zc = zc * (1 / (b - a) * (
                        0.5 * (a - xc) ** 2 * np.log(np.abs(a - xc))
                        - 0.5 * (b - xc) ** 2 * np.log(np.abs(b - xc))
                        + 0.25 * (b - xc) ** 2
                        - 0.25 * (a - xc) ** 2
                ) - xc * np.log(xc) + g - h * xc)
                return zc

        return np.vectorize(f_zc)


class NACA16Series(BaseNACA):
    """
    References:
        -   NACA Report No.824, Summary of Aerofoil Data

    """
    _pattern_valid = r"^16-\d{3}((,a=0?(\.\d+)?)|(,a=1(\.0+)?))?$"

    # noinspection PyProtectedMember
    _f_nd_yz = NACA6Series._f_nd_yz  # Identical camber function
    # noinspection PyProtectedMember
    _f_camber = NACA6Series._f_camber
    # noinspection PyProtectedMember
    _f_nd_yt = NACA4DigitModifiedSeries._f_nd_yt  # Identical thickness function

    def __init__(self, code: str):
        """
        Args:
            code: NACA aerofoil code.
        """
        # Superclass call
        BaseNACA.__init__(self, code=code)

        # Parse
        code, _ = self._pop_pattern(code, r"\d")
        code, p_min_x = self._pop_pattern(code, r"\d")
        code, _ = self._pop_pattern(code, "-")
        code, cli = self._pop_pattern(code, r"\d")
        code, t_max = self._pop_pattern(code, r"\d{2}")
        # Meanline
        if re.match("^$", code):
            a = 1
        else:
            code, _ = self._pop_pattern(code, ",a=")
            a = code

        # Recast
        self._parsed.p_min_x = float(p_min_x) / 10
        self._parsed.meanline_cli = [float(cli) / 10]
        self._parsed.t_max = float(t_max) / 100
        self._parsed.meanline_a = [float(a)]
        # Secret definitions, implied by nature of 16-Series
        self._parsed.LEri = 4
        self._parsed.t_max_x = 0.5
        return



"""Module implementing equations of state for gases."""
import numpy as np
import re

from carpy.utility import Quantity, cast2numpy, constants as co
from carpy.gaskinetics._chemistry_molecules import MoleculeAcyclic

__all__ = ["GasModels"]
__author__ = "Yaseen Reza"

# Redesignate
Molecule = MoleculeAcyclic


# ============================================================================ #
# Thermodynamic state variables, Gas mixing
# ---------------------------------------------------------------------------- #
class IntensiveProperties(object):
    """Data class for tracking intensive properties of thermodynamic state."""

    _rho = Quantity(np.nan, "kg m^{-3}")
    _cv = Quantity(np.nan, "J kg^{-1} K^{-1}")
    _u = Quantity(np.nan, "J kg^{-1}")
    _p = Quantity(np.nan, "Pa")
    _T = Quantity(np.nan, "K")

    # -------------------- #
    # Intensive properties #
    # -------------------- #

    @property
    def a(self) -> Quantity:
        """Chemical thermodynamic activity."""
        return NotImplemented

    @property
    def mu(self) -> Quantity:
        """Chemical potential."""
        return NotImplemented

    @property
    def betaT(self) -> Quantity:
        """Isothermal compressibility."""
        return NotImplemented

    @property
    def betaS(self) -> Quantity:
        """Adiabatic compressibility."""
        return NotImplemented

    @property
    def Kf(self) -> Quantity:
        """Cryoscopic constant."""
        return NotImplemented

    @property
    def rho(self) -> Quantity:
        """Density."""
        return self._rho

    @property
    def Kb(self) -> Quantity:
        """Ebullioscopic."""
        return NotImplemented

    @property
    def h(self) -> Quantity:
        """Specific enthalpy."""
        return NotImplemented

    @property
    def s(self) -> Quantity:
        """Specific entropy."""
        # Probably have to work out molecular inertia, and then use:
        # https://doi.org/10.3390/e21050454
        return NotImplemented

    @property
    def f(self) -> Quantity:
        """Fugacity."""
        return NotImplemented

    @property
    def g(self) -> Quantity:
        """Specific Gibbs free energy."""
        return NotImplemented

    @property
    def cp(self) -> Quantity:
        """Specific heat capacity at constant pressure (isobaric)."""
        return NotImplemented

    @property
    def cv(self) -> Quantity:
        """Specific heat capacity at constant volume (isochoric)."""
        return self._cv

    @property
    def u(self) -> Quantity:
        """Specific internal energy."""
        return self._u

    @property
    def piT(self) -> Quantity:
        """Internal pressure."""
        return NotImplemented

    @property
    def p(self) -> Quantity:
        """Pressure."""
        return self._p

    @property
    def T(self) -> Quantity:
        """Temperature."""
        return self._T

    @property
    def k_thermal(self) -> Quantity:
        """Thermal conductivity."""
        return NotImplemented

    @property
    def alpha(self) -> Quantity:
        """Thermal diffusivity."""
        return NotImplemented

    @property
    def alphaV(self) -> Quantity:
        """Volumetric coefficient of thermal expansion at constant pressure."""
        return NotImplemented

    @property
    def Chi(self) -> Quantity:
        """Vapour quality."""
        return NotImplemented

    @property
    def nu(self) -> Quantity:
        """Specific volume."""
        return 1 / self.rho

    @property
    def gamma(self) -> float:
        """Specific heat ratio (a.k.a heat capacity ratio, adiabatic index)."""
        return float(self.cp / self.cv)


class StateVars(IntensiveProperties):
    """
    Data class for tracking intensive, extensive, and other derived properties
    of thermodynamic state in a system.
    """
    _n: Quantity

    # -------------------- #
    # Extensive properties #
    # -------------------- #

    @property
    def H(self) -> Quantity:
        """Enthalpy."""
        return self.m * self.h

    @property
    def S(self) -> Quantity:
        """Entropy."""
        return self.m * self.s

    @property
    def G(self) -> Quantity:
        """Gibbs free energy."""
        return self.m * self.g

    @property
    def Xi(self) -> Quantity:
        """Planck potential, Gibbs free entropy."""
        return self.Phi - self.p * self.V / self.T

    @property
    def Ohm(self) -> Quantity:
        """Landau potential, Landau free energy."""
        return self.U - self.T * self.S - self.mu * self.N

    @property
    def Cp(self) -> Quantity:
        """Heat capacity at constant pressure (isobaric)."""
        return self.m * self.cp

    @property
    def Cv(self) -> Quantity:
        """Heat capacity at constant volume (isochoric)."""
        return self.m * self.cv

    @property
    def F(self) -> Quantity:
        """Helmholtz free energy."""
        return self.U - self.T * self.S

    @property
    def Phi(self) -> Quantity:
        """Massieu potential, Helmholtz free entropy."""
        return self.S - self.U / self.T

    @property
    def U(self) -> Quantity:
        """Internal energy."""
        return self.m * self.u

    @property
    def m(self) -> Quantity:
        """Mass."""
        return NotImplemented

    @property
    def N(self) -> int:
        """Number of particles."""
        return int(self.n * co.PHYSICAL.N_A)

    @property
    def V(self) -> Quantity:
        """Volume."""
        return self.m * self.nu

    @property
    def n(self) -> Quantity:
        """Molar mount of substance."""
        return self._n


class PerfectGases(StateVars):
    """
    A class for modelling mixtures of ideal (perfect or semi-perfect) gases.

    Examples: ::

        # Create a gas approximately representative of the composition of air
        >>> air = PerfectGases().X = "N2:78.0, O2:21.0"

    """
    _X: dict[Molecule, float]  # Template for what composition should look like
    _n = Quantity(1, "mol")  # Define an amount of substance, just for the lulz

    def __init__(self, name: str = None):
        """
        Args:
            name: For your own convenience, a unique identifier for this object.
        """
        self._name = name
        return

    def __repr__(self):
        reprstr = f"{type(self).__name__}(name='{self.name}')"
        return reprstr

    @property
    def name(self) -> str:
        """Human-readable name of the gas composition."""
        if self._name is None:
            return 'unnamed-gas'
        return self._name

    @staticmethod
    def _XYparse(value) -> dict:
        # User has set value as a dictionary
        if isinstance(value, dict):
            out = dict()
            # Make sure dictionary entries are molecule or string objects
            for k, v in value.items():
                if isinstance(k, Molecule):
                    out[k] = float(v)
                elif isinstance(k, str):
                    out[Molecule(k)] = float(v)
                else:
                    errormsg = (
                        f"Expected one type of either 'str' or "
                        f"{Molecule.__name__}, got type={type(k)} from "
                        f"object {k.__repr__()}"
                    )
                    raise TypeError(errormsg)

        # User has set value as a string
        elif isinstance(value, str):
            out = dict([
                (Molecule(formula.replace(" ", "")), float(v))
                for (formula, v) in re.findall(r"(.+?):(\d*\.?\d*),?\s*", value)
            ])
            # Check if user accidentally omitted mole amount on single molecule
            if len(out) == 0:
                try:
                    out = dict([(Molecule(value.replace(" ", "")), 1.0)])
                except Exception:
                    errormsg = f"Couldn't parse gas components in '{value}'"
                    raise ValueError(errormsg)
            # If mole amounts are given, molecules should match number of colons
            elif len(out) < value.count(":"):
                raise ValueError(f"Lossy conversion from '{value}' to '{out}'")

        # Invalid argument
        else:
            errormsg = f"{value} is not a valid definition of gas composition"
            raise ValueError(errormsg)

        # The compositional dictionary is valid, now normalise fraction to unity
        vector_magnitude = sum(out.values())
        out = {k: (v / vector_magnitude) for (k, v) in out.items()}
        return out

    @property
    def X(self) -> dict[Molecule, float]:
        """Composition of gas, by mole fraction X."""
        return self._X

    @X.setter
    def X(self, value):
        self._X = self._XYparse(value=value)
        return

    @property
    def Y(self) -> dict[Molecule, float]:
        """Composition of gas, by mass fraction Y."""
        X_i = cast2numpy(list(self.X.values()))
        M_i = cast2numpy([float(molecule.M) for molecule in self.X.keys()])
        M = np.sum(X_i * M_i)
        Y_i = (X_i * (M_i / M))
        Y = dict(zip(self.X.keys(), Y_i))
        return Y

    @Y.setter
    def Y(self, value):
        Y = self._XYparse(value=value)
        Y_i = cast2numpy(list(Y.values()))
        M_i = cast2numpy([float(molecule.M) for molecule in Y.keys()])
        M = 1 / np.sum(Y_i / M_i)
        X_i = (Y_i / (M_i / M))
        self._X = dict(zip(Y.keys(), X_i))
        return

    @property
    def _M_components(self) -> Quantity:
        """Molar mass of the gas mixture, by gas."""
        X_i = cast2numpy(list(self.X.values()))
        M_i = Quantity([key.M[0] for key in self.X.keys()], units="kg mol^{-1}")
        M_i = (X_i * M_i)
        return M_i

    @property
    def M(self) -> Quantity:
        """Average molar mass of the gas mixture."""
        M = self._M_components.sum()
        return M

    @property
    def m(self) -> Quantity:
        """Mass of gas in the mixture."""
        return self.n * self.M

    @property
    def cp(self) -> Quantity:
        """Specific heat capacity at constant pressure (isobaric)."""
        return self.cv + co.PHYSICAL.R / self.M

    @property
    def Rs(self) -> Quantity:
        """Specific gas constant."""
        Rs = co.PHYSICAL.R / self.M
        return Rs

    @property
    def c_sound(self) -> Quantity:
        """Speed of sound in an ideal gas."""
        # Isentropic bulk modulus (Ks)
        Ks = self.gamma * self.p
        c_sound = (Ks / self.rho) ** 0.5
        return c_sound


# ============================================================================ #
# Gas Modelling with Equations of State
# ---------------------------------------------------------------------------- #
class GasModel(object):
    """
    object: "What is my purpose?"
    yaseen: "You are a typehint."
    object: "oh my god."
    """
    pass


class PerfectCaloric(GasModel, PerfectGases):
    """
    Equation of state model for an ideal, calorically perfect gas.

    Gases modelled as calorically perfect have specific heat capacities that are
    independent of temperature.
    """

    def __init__(self, name: str = None):
        # Superclass call
        PerfectGases.__init__(self, name=name)
        return

    def __copy__(self):
        # Shallow copy - take name and state
        copy = type(self)(name=self._name)
        copy.X = self.X
        copy.state_TP = self.state_TP
        return copy

    def copy(self):
        """Return a shallow copy of the object (a new gas object)."""
        return self.__copy__()

    @property
    def _cv_components(self) -> Quantity:
        """Specific heat capacity at constant volume (isochoric), by gas."""
        molecule_dof = np.zeros(len(self.X))
        for i, molecule in enumerate(self.X.keys()):
            axe_all = molecule.atom_AXE
            dof = 0

            # Translational DoF
            dof += 3
            if len(axe_all) == 0:
                molecule_dof[i] = dof
                continue  # There were no bonds to any other atoms/ligands, skip

            # Rotational DoF
            axe_central = {
                atom: xe for (atom, xe) in axe_all.items()
                if len(xe["X"]) > 1
            }
            linear = all([
                len(v["X"]) == 2 and v["E"] == 0 for v in axe_central.values()])
            dof += 2
            if linear is False:
                dof += 1

            molecule_dof[i] = dof
            continue  # Other degrees of freedom are not considered

        # Compute isochoric heat capacity
        Xi = cast2numpy(list(self.X.values()))
        Yi = cast2numpy(list(self.Y.values()))
        Mi = (Yi / Xi) * self.M
        Ri = Yi * (co.PHYSICAL.R / Mi)
        cv_i = ((molecule_dof / 2) * Ri)
        return cv_i

    @property
    def _cv(self) -> Quantity:
        """Specific heat capacity at constant volume (isochoric)."""
        return self._cv_components.sum()

    @property
    def _u_components(self) -> Quantity:
        """Specific internal energy, by gas."""
        return self._cv_components * self.T

    @property
    def _u(self) -> Quantity:
        """Specific internal energy."""
        return self._u_components.sum()

    @property
    def h(self) -> Quantity:
        """Specific enthalpy."""
        return self.u + self.p * self.nu

    @property
    def state_TP(self) -> tuple[Quantity, Quantity]:
        """Thermodynamic state temperature and pressure variables."""
        return self.T, self.p

    @state_TP.setter
    def state_TP(self, value):
        # Unpack value and recast
        T, p = value
        T = self._T if T is None else Quantity(T, "K")
        p = self._p if p is None else Quantity(p, "Pa")

        # Compute missing parameters
        rho = p / self.Rs / T

        # Make assignments
        self._rho = rho
        self._p = p
        self._T = T
        return

    @property
    def state_TD(self) -> tuple[Quantity, Quantity]:
        """Thermodynamic state temperature and density variables."""
        return self.T, self.rho

    @state_TD.setter
    def state_TD(self, value):
        # Unpack value and recast
        T, rho = value
        T = self._T if T is None else Quantity(T, "K")
        rho = self._rho if rho is None else Quantity(rho, "kg m^{-3}")

        # Compute missing parameters
        p = rho * self.Rs * T

        # Make assignments
        self._rho = rho
        self._p = p
        self._T = T
        return


class PerfectThermic(GasModel, PerfectGases):
    """
    Equation of state model for an ideal, thermally perfect gas.

    Gases modelled as thermally perfect have specific heat capacities that are
    functions of temperature alone.
    """

    def __init__(self, name: str = None):
        # Superclass call
        PerfectGases.__init__(self, name=name)
        raise NotImplementedError("Sorry! Use 'PerfectCaloric' instead")


class GasModels(object):
    """A collection of gas models, packaged for easy access."""

    # Model typehint
    typehint = GasModel

    # Models
    PerfectCaloric = PerfectCaloric


# Destructive instantiation (no one will want the class of this anyway)
GasModels = GasModels()

# # ============================================================================ #
# # Gas Modelling with Equations of State
# # ---------------------------------------------------------------------------- #
# # ---------------------------------------------------------------------------- #
# class ModelRK(BaseGas):
#     """
#     Real gas model based on the Redlich-Kwong equation of state.
#     """
#
#     def __init__(self, formula: str):
#         """
#         Args:
#             formula: Condensed structural formula of the gas being modelled.
#         """
#         super().__init__(formula)
#         return
#
#     def _f_Z(self, T: np.ndarray, p: np.ndarray):
#         # Recast as necessary
#         T, p = np.broadcast_arrays(T, p)
#         critT, critp = self.molecule.critical_point
#         if np.isnan(critT) or np.isnan(critp):
#             raise ValueError("Couldn't find data for critical point.")
#         else:
#             critT = float(critT)
#             critp = float(critp)
#         R = co.PHYSICAL.R.x
#
#         # Molecule-dependent quantities
#         croot2 = 2 ** (1 / 3)
#         a = 1 / (9 * (croot2 - 1)) * R ** 2 * critT ** 2.5 / critp
#         b = (croot2 - 1) / 3 * R * critT / critp
#
#         def solver(temperature: float, pressure: float):
#             """Solve for compressibility coefficient given scalar quantities."""
#
#             # Solve for the molar volume that makes the whole expression work
#             def f_opt(x):
#                 """Equation solving for the molar volume (V/n)."""
#                 lhs = pressure * x / R / temperature
#                 rhs = x / (x - b) - a / (R * temperature ** 1.5 * (x + b))
#                 return lhs - rhs
#
#             V_m = newton(f_opt, 0.024)
#             Z = pressure * V_m / R / temperature
#             return Z
#
#         return np.vectorize(solver)(T, p)
